import io
import os
import csv
import subprocess

import sdi_utils.gensolution as gs
import deprecated.set_logging as slog
import sdi_utils.textfield_parser as tfp
import sdi_utils.tprogress as tp

try:
    api
except NameError:
    class api:
        class Message:
            def __init__(self,body = None,attributes = ""):
                self.body = body
                self.attributes = attributes
                
        def send(port,msg) :
            if port == outports[1]['name'] :
                print(msg.attributes)
                for i,dk in enumerate(msg.body) :
                    print(dk)
                    if i > 100 :
                        break
                print(msg.body)
            else :
                print('{}: {}'.format(port,msg))
    
        def call(config,csvstream):
            api.config = config
            return process(csvstream)
            
        def set_port_callback(port, callback) :
            pass
    
        class config:
            ## Meta data
            config_params = dict()
            tags = {'sdi_utils':''}
            version = "0.0.1"
            operator_name = 'csv_dict'
            operator_description = "csv stream to dict"
            operator_description_long = "Converts csv stream to dict"
            add_readme = dict()
            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}
            collect = True
            config_params['collect'] = {'title': 'Collect data', 'description': 'Collect data before sending it to the output port',
                                           'type': 'boolean'}
            separator = ';'
            config_params['separator'] = {'title': 'Separator', 'description': 'Separator',
                                           'type': 'string'}

            column_dict = 'key_col_header:value_col_header'
            config_params['column_dict'] = {'title': 'Column dictionary', 'description': 'Column dictionary with first \
with first column as key and second column as value','type': 'string'}

result_dicts = list()

def process(msg):
    att_dict = msg.attributes

    global result_dicts

    att_dict['operator'] = 'csv_dict'
    logger, log_stream = slog.set_logging(att_dict['operator'], loglevel=api.config.debug_mode)
    logger.info("Process started")
    time_monitor = tp.progress()

    logger.debug('Attributes: {}'.format(str(att_dict)))

    str_decode = msg.body.decode('utf-8')
    csv_io = io.StringIO(str_decode)
    dictreader = csv.DictReader(csv_io, delimiter=api.config.separator)

    col_dict = tfp.read_dict(api.config.column_dict)
    if col_dict :
        if api.config.collect :
            error_msg = 'Error: parameter <collect=True> and column dict are not implemented. Choose either-or.'
            logger.error(error_msg)
            raise ValueError(error_msg)
        logger.info("Create dict with first column as key and second as value")
        for key_col,val_col in col_dict.items() :
            cdict = {}
            for row in dictreader:
                for header, value in row.items():
                    try:
                        cdict[header].append(value)
                    except KeyError:
                        cdict[header] = [value]
            result_dicts.extend(cdict)
    else:
        logger.info("Create list of dictionaries for each row")
        dict_list = [x for x in dictreader]
        result_dicts.extend(dict_list)

    logger.debug('Process ended: {}'.format(time_monitor.elapsed_time()))
    if msg.attributes['message.lastBatch'] or api.config.collect == False:
        msg = api.Message(attributes=att_dict, body=result_dicts)
        api.send(outports[1]['name'], msg)

    if api.config.debug_mode:
        for i, dk in enumerate(result_dicts):
            logger.debug(dk)
            if i > 100:
                break

    api.send(outports[0]['name'], log_stream.getvalue())


inports = [{'name': 'stream', 'type': 'message.file',"description":"Input file message"}]
outports = [{'name': 'log', 'type': 'string',"description":"Logging data"}, \
            {'name': 'data', 'type': 'message.dicts',"description":"Output data as list of dictionaries"}]


#api.set_port_callback(inports[0]['name'], process)

def test_operator() :
    config = api.config
    config.debug_mode = True
    config.collect = False
    config.separator = ','
    config.column_dict = 'index:keyword'

    fname = '/Users/Shared/data/onlinemedia/repository/lexicon.csv'

    fbase = fname.split('.')[0]
    attributes = {'format': 'csv', "storage.filename": fbase, 'message.lastBatch': True, \
                  'storage.fileIndex': 0, 'storage.fileCount': 1,'process_list':[]}
    csvstream = open(fname, mode='rb').read()
    msg = api.Message(attributes=attributes, body=csvstream)
    api.call(config=config, csvstream = msg)


if __name__ == '__main__':
    #test_operator()
    if True :
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name+'_'+api.config.version

        subprocess.run(["vctl", "solution", "bundle", '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_0.0.1',\
                                  "-t", solution_name])
        subprocess.run(["mv", solution_name+'.zip', '../../../solution/operators'])
        



import sdi_utils.gensolution as gs
import deprecated.set_logging as slog
import sdi_utils.textfield_parser as tfp

import requests
import subprocess
import os
import csv


try:
    api
except NameError:
    class api:
        class Message:
            def __init__(self,body = None,attributes = ""):
                self.body = body
                self.attributes = attributes
                
        def send(port,msg) :
            if outports[1]['name'] == port :
                print(msg.attributes)
                with open('/Users/Shared/data/test/rki.csv', 'w') as f:
                    writer = csv.writer(f)


        def set_config(config):
            api.config = config

        class config:
            ## Meta data
            config_params = dict()
            version = '0.0.1'
            tags = {'sdi_utils':'', 'requests':''}
            operator_name = 'HTTPdownload'
            operator_description = "HTTP File Download"
            operator_description_long = "HTTP File Download."
            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}
            url = 'None'
            config_params['url'] = {'title': 'URL of file', 'description': 'URL of file to download', 'type': 'string'}


def process() :
    att_dict = {}
    att_dict['operator'] = 'HTTPdownload'
    logger, log_stream = slog.set_logging(att_dict['operator'])
    if api.config.debug_mode == True:
        logger.setLevel('DEBUG')

    logger.info("Process started")

    url_file = tfp.read_value(api.config.url)

    r = requests.get(url_file)
    if r.status_code != 200 :
        logger.warning('File not found: {}. \'None\' send to outport.'.format(url_file))
        downloaded_file = None
    else :
        downloaded_file = r.content

    att_dict["file"] = {"connection": "", "path": ""}
    msg = api.Message(attributes=att_dict, body=downloaded_file)
    api.send(outports[1]['name'], msg)
    logger.info('File send to outport: {}'.format(url_file))
    logger.info("Process ended")
    api.send(outports[0]['name'],log_stream.getvalue())

inports = []
outports = [{"name":"log","type":"string","description":"Logging"},\
            {"name":"csv","type":"message.file","description":"url file"}]

#api.add_generator(process)

def test_operator() :
    config = api.config
    config.debug_mode = True
    config.url = "https://opendata.arcgis.com/datasets/dd4580c810204019a7b8eb3e0b329dd6_0.csv"
    api.set_config(config)
    process()



if __name__ == '__main__':
    #test_operator()
    #gs.gensolution(realpath(__file__), api.config, None, outports)
    if True :
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name+'_'+api.config.version
        print('Solution name: {}'.format(solution_name))
        print('Current directory: {}'.format(os.getcwd()))
        subprocess.run(["vctl", "solution", "bundle", '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_0.0.1',\
                                  "-t", solution_name])
        subprocess.run(["mv", solution_name+'.zip', '../../../solution/operators'])


        



import os
import pandas as pd

import subprocess

import sdi_utils.gensolution as gs
import deprecated.set_logging as slog
import sdi_utils.textfield_parser as tfp

try:
    api
except NameError:
    class api:
        class Message:
            def __init__(self,body = None,attributes = ""):
                self.body = body
                self.attributes = attributes
                
        def send(port,msg) :
            if port == outports[1]['name'] or port == outports[2]['name'] :
                print(msg.attributes)
                print(msg.body)

        class config:
            ## Meta data
            config_params = dict()
            tags = {'sdi_utils':'', 'pandas':''}
            version = "0.1.0"

            operator_description = "dict to df"
            operator_name = 'dict_df'
            operator_description_long = "Converts dict to DataFrame"
            add_readme = dict()
            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}

            rename = 'None'
            config_params['rename'] = {'title': 'Rename columns',
                                           'description': 'Rename columns by provided map',
                                           'type': 'string'}

            send_empty = True
            config_params['send_empty'] = {'title': 'Send empty Dataframe to \'outport\'.',
                                           'description': 'Send empty Dataframe to outport or status to \'error port\'.',
                                           'type': 'boolean'}




def process(msg) :
    att_dict = msg.attributes

    att_dict['operator'] = 'dict_df'
    logger, log_stream = slog.set_logging(att_dict['operator'], loglevel=api.config.debug_mode)
    logger.info("Process started. Logging level: {}".format(logger.level))

    logger.debug('Attributes: {}'.format(str(msg.attributes)))
    if msg.body == None  or len(msg.body) == 0 :
        logger.warning('Empty dictionary received. Empty DataFrame will be send.')

    df = pd.DataFrame(msg.body)

    rename_cols = tfp.read_dict(api.config.rename)
    if rename_cols :
        df.rename(columns = rename_cols,inplace=True)

    logger.debug('DataFrame Shape: {} - {}'.format(df.shape[0],df.shape[1]))


    att_dict['data_outcome'] = '0' if df.shape[0] == 0 else '1'

    if df.shape[0] == 0 & api.config.send_empty == False :
        api.send(outports[2]['name'],api.Message(attributes=att_dict,body = att_dict['data_outcome']))
    else :
        api.send(outports[1]['name'], api.Message(attributes=att_dict,body=df))

    api.send(outports[0]['name'], log_stream.getvalue())


inports = [{'name': 'dict', 'type': 'message',"description":"Input dict"}]
outports = [{'name': 'log', 'type': 'string',"description":"Logging data"}, \
            {'name': 'df', 'type': 'message.DataFrame',"description":"Output df"},
            {'name': 'error', 'type': 'message',"description":"Error status"}]


#api.set_port_callback(inports[0]['name'], process)

def test_operator() :

    api.config.debug_mode = True
    api.config.send_empty = False
    #api.config.rename = "name: first_name, age: AGE"


    data = [{'name': 'Anna', 'age': 25, 'born': '1995-03-14'}, {'name': 'Berta', 'age': 31, 'born': '1989-08-14'},
            {'name': 'Cecilia', 'age': 41, 'born': '1979-08-14'}]
    msg = api.Message(attributes={'format': 'json'}, body=data)
    process(msg)

    msg = api.Message(attributes={'format': 'json'}, body={})
    process(msg)

if __name__ == '__main__':
    #test_operator()
    if True:
        subprocess.run(["rm", '-r',
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version])
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name + '_' + api.config.version
        subprocess.run(["vctl", "solution", "bundle",
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version, \
                        "-t", solution_name])
        subprocess.run(["mv", solution_name + '.zip', '../../../solution/operators'])






import sdi_utils.gensolution as gs
import deprecated.set_logging as slog
import sdi_utils.textfield_parser as tfp
import sdi_utils.tprogress as tp


import subprocess
import os
import pandas as pd



try:
    api
except NameError:
    class api:

        queue = list()
        class Message:
            def __init__(self,body = None,attributes = ""):
                self.body = body
                self.attributes = attributes
                
        def send(port,msg) :
            if port == outports[1]['name'] :
                api.queue.append(msg)
    
        def set_config(config):
            api.config = config

        class config:
            ## Meta data
            config_params = dict()
            version = '0.0.1'
            tags = {'pandas': '','sdi_utils':''}
            operator_name = 'drop_duplicates'
            operator_description = "Drop duplicates of Dataframe"
            operator_description_long = "Drop duplicates of Dataframe based on subset"
            add_readme = dict()
            add_readme["References"] = r"""[pandas doc: to_csv](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html)"""

            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}

            subset = 'None'
            config_params['subset'] = {'title': 'Subset of columns','description': 'Subset of columns for duplicates check.','type': 'string'}

            keyword_args = "None"
            config_params['keyword_args'] = {'title': 'Keyword Arguments',
                                             'description': 'Mapping of key-values passed as arguments \"to read_csv\"',
                                             'type': 'string'}

def process(msg) :

    att_dict = msg.attributes

    att_dict['operator'] = 'drop_duplicates'
    logger, log_stream = slog.set_logging(att_dict['operator'], loglevel=api.config.debug_mode)
    logger.info("Process started. Logging level: {}".format(logger.level))
    time_monitor = tp.progress()
    logger.debug('Attributes: {}'.format(str(msg.attributes)))

    # start custom process definition
    df = msg.body

    subset = tfp.read_list(api.config.subset)
    kwargs = tfp.read_dict(text=api.config.keyword_args, map_sep='=')
    if subset :
        if not kwargs == None :
            df.drop_duplicates(subset = subset,inplace=True, **kwargs)
        else :
            df.drop_duplicates(subset = subset,inplace=True)

    logger.debug('Process ended: {}'.format(time_monitor.elapsed_time()))

    api.send(outports[0]['name'],log_stream.getvalue())
    api.send(outports[1]['name'],api.Message(attributes=att_dict,body = df))



inports = [{'name': 'data', 'type': 'message.DataFrame',"description":"Input data"}]
outports = [{'name': 'log', 'type': 'string',"description":"Logging data"}, \
            {'name': 'data', 'type': 'message.DataFrame',"description":"Output data"}]



#api.set_port_callback(inports[0]['name'], process)

def test_operator() :

    config = api.config
    config.debug_mode = True
    config.subset = "icol, 'col 2'"
    api.set_config(config)

    df = pd.DataFrame({'icol': [1, 2, 1, 4, 5], 'col 2': ['2020-01-01', '2020-02-01', '2020-01-01', '2020-01-28','2020-04-12'],\
                       'col3': [100.0, 200.2, 300.4, 400, 500],'names':['Anna','Berta','Berta','Claire','Dora'],\
                       'bool':[True, False, False, True, True]})

    attributes = {'format': 'csv', 'name': 'DF_name','process_list':[]}
    msg = api.Message(attributes=attributes,body=df)
    process(msg)

    df_list = [d.body for d in api.queue]
    print(pd.concat(df_list))

if __name__ == '__main__':
    #test_operator()
    if True :
        subprocess.run(["rm", '-r',
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators' + api.config.version])
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name + '_' + api.config.version
        subprocess.run(["vctl", "solution", "bundle",
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version, \
                        "-t", solution_name])
        subprocess.run(["mv", solution_name + '.zip', '../../../solution/operators'])
        



import sdi_utils.gensolution as gs
import deprecated.set_logging as slog
import sdi_utils.textfield_parser as tfp
import sdi_utils.tprogress as tp

import subprocess
import os
import pandas as pd

try:
    api
except NameError:
    class api:

        queue = list()

        class Message:
            def __init__(self, body=None, attributes=""):
                self.body = body
                self.attributes = attributes

        def send(port, msg):
            if port == outports[1]['name']:
                api.queue.append(msg)

        class config:
            ## Meta data
            config_params = dict()
            version = '0.0.1'
            tags = {'pandas': '', 'sdi_utils': ''}
            operator_name = 'df_csv'
            operator_description = "df to csv"
            operator_description_long = "Creates a csv-formatted data passed to outport as message with the csv-string as body."
            add_readme = dict()
            add_readme[
                "References"] = r"""[pandas doc: to_csv](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_csv.html)"""

            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}
            write_index = False
            config_params['write_index'] = {'title': 'Write Index', 'description': 'Write index or ignore it',
                                            'type': 'boolean'}

            separator = ','
            config_params['separator'] = {'title': 'Separator', 'description': 'Separator', 'type': 'string'}

            reset_index = False
            config_params['reset_index'] = {'title': 'Reset Index', 'description': 'Reset index or indices',
                                            'type': 'boolean'}

            rename = 'None'
            config_params['rename'] = {'title': 'Rename Columns', 'description': 'Rename columns (map)',
                                       'type': 'string'}

            select_columns = 'None'
            config_params['select_columns'] = {'title': 'Select Columns', 'description': 'Select columns.',
                                               'type': 'string'}
            datetime_conversion = False
            config_params['datetime_conversion'] = {'title': 'Datetime conversion (YYYY:MM:DD hh:mm:ss)',
                                            'description': 'Converting Datetime columns to format: YYYY:MM:DD hh:mm:ss.',
                                            'type': 'boolean'}

            bool_to_int = True
            config_params['bool_to_int'] = {'title': 'Convert boolean to int',
                                            'description': 'Converting boolean value to integer.',
                                            'type': 'boolean'}

            keyword_args = "None"
            config_params['keyword_args'] = {'title': 'Keyword Arguments',
                                             'description': 'Mapping of key-values passed as arguments \"to read_csv\"',
                                             'type': 'string'}

            sort_columns = False
            config_params['sort_columns'] = {'title': 'Sort Columns alphabetically',
                                             'description': 'Sort Columns alphabetically',
                                             'type': 'boolean'}


def process(msg):
    att_dict = msg.attributes

    att_dict['operator'] = 'df_csv'
    logger, log_stream = slog.set_logging(att_dict['operator'], loglevel=api.config.debug_mode)
    logger.info("Process started. Logging level: {}".format(logger.level))
    time_monitor = tp.progress()
    logger.debug('Attributes: {}'.format(str(msg.attributes)))

    df = msg.body

    if not isinstance(df, pd.DataFrame):
        logger.warning('Body of message is not of type DataFrame. No processing')
        logger.debug('Body value: {}'.format(msg.body))
        api.send(outports[0]['name'], log_stream.getvalue())
        api.send(outports[1]['name'], api.Message(attributes=att_dict, body=df))
        return 0
    if df.empty:
        logger.warning('DataFrame is empty. No processing')
        api.send(outports[0]['name'], log_stream.getvalue())
        api.send(outports[1]['name'], api.Message(attributes=att_dict, body=df))
        return 0

    logger.debug('DataFrame shape: {} - {}'.format(df.shape[0], df.shape[1]))

    data_str = ''

    try:
        # start custom process definition
        if api.config.reset_index:
            logger.debug('Reset Index')
            df = df.reset_index()

        rename_dict = tfp.read_dict(api.config.rename)
        if rename_dict:
            prev_columns = df.columns
            logger.debug('Renaming columns: {} ({} - {})'.format(rename_dict, prev_columns, df.columns))
            df.rename(columns=rename_dict, inplace=True)

        # Datetime
        if api.config.datetime_conversion :
            dt_cols = df.select_dtypes(include=['datetime64[ns, UTC]', 'datetime64[ns]', 'datetime64']).columns
            logger.debug('Datetime columns: {}'.format(dt_cols))
            for col in dt_cols:
                logger.debug('Column datetime conversion: {}'.format(col))
                df[col] = df[col].dt.strftime('%Y-%m-%d %H:%M:%S')

        select_columns = tfp.read_list(api.config.select_columns)
        if select_columns:
            logger.debug('Column selection: {}'.format(select_columns))
            df = df[select_columns]

        if api.config.bool_to_int :
            bool_cols = df.select_dtypes(include='bool').columns
            logger.debug('Boolean columns: {}'.format(bool_cols))
            for bc in bool_cols:
                logger.debug('Boolean column conversion to int: {}'.format(bc))
                df[bc] = df[bc].astype(int)

        sep = api.config.separator
        if not sep:
            sep = ','
        logger.debug('to_csv - delimiter: {}'.format(sep))

        logger.debug('to_cvs - write_index: {}'.format(api.config.write_index))

        #df.dropna(inplace=True)
        logger.debug('DataFrame shape: {} - {}'.format(df.shape[0], df.shape[1]))

        # to ensure that the df has always the same order
        # when reading from DB this can sometimes change
        if api.config.sort_columns :
            df = df.reindex(sorted(df.columns), axis=1)

        kwargs = tfp.read_dict(text=api.config.keyword_args, map_sep='=')

        if not kwargs == None:
            logger.debug('to_csv - kwargs: {}'.format(kwargs))
            data_str = df.to_csv(sep=sep, index=api.config.write_index, **kwargs)
        else:
            logger.debug('to_csv - no kwargs.')
            data_str = df.to_csv(sep=sep, index=api.config.write_index)

    except Exception as e:
        logger.error(e)
        logger.debug('Process terminated with Error: {}'.format(time_monitor.elapsed_time()))
        api.send(outports[0]['name'], log_stream.getvalue())


    logger.debug('Process ended: {}'.format(time_monitor.elapsed_time()))
    logger.debug('Datastr: {}'.format(data_str))
    api.send(outports[0]['name'], log_stream.getvalue())
    api.send(outports[1]['name'], api.Message(attributes=att_dict, body=data_str))


inports = [{'name': 'data', 'type': 'message.DataFrame', "description": "Input data"}]
outports = [{'name': 'log', 'type': 'string', "description": "Logging data"}, \
            {'name': 'csv', 'type': 'message', "description": "Output data as csv"}]

#api.set_port_callback(inports[0]['name'], process)


def test_operator():

    api.config.write_index = False
    api.config.reset_index = True
    api.config.rename = 'icol:index, col3: column3'
    api.config.select_columns = "index, 'col 2', names, bool"
    api.config.bool_to_int = True
    api.config.sort_columns = True

    df = pd.DataFrame(
        {'icol': [1, 2, 3, 4, 5], 'col 2': ['2020-01-01', '2020-02-01', '2020-01-31', '2020-01-28', '2020-04-12'], \
         'col3': [100.0, 200.2, 300.4, 400, 500], 'names': ['Anna', 'Berta', 'Berta', 'Claire', 'Dora'], \
         'bool': [True, False, False, True, True]})
    df = df.set_index(keys=['icol'])
    df['col 2'] = pd.to_datetime(df['col 2'], format='%Y-%m-%d', utc=True)

    attributes = {'format': 'csv', 'name': 'DF_name', 'process_list': []}
    msg = api.Message(attributes=attributes, body=df)
    process(msg)

    str_list = [d.body for d in api.queue]
    str_str = '\n'.join(str_list)
    print(str_str)
    # out_file = '/Users/Shared/data/test/json_df.csv'
    # df.to_csv(out_file,index=False)

if __name__ == '__main__':
    #test_operator()
    if True:
        subprocess.run(["rm", '-r',
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version])
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name + '_' + api.config.version
        subprocess.run(["vctl", "solution", "bundle",
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version, \
                        "-t", solution_name])
        subprocess.run(["mv", solution_name + '.zip', '../../../solution/operators'])






import subprocess
import sys
import io
import select
import os


def main():

    cmd = ["python", "outputgen.py"]
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,stderr=subprocess.PIPE,universal_newlines=True)

    output = '1'
    errout = '1'
    while proc.poll() is None or  output != '' or  errout != '':
        output = proc.stdout.readline()
        errout = proc.stderr.readline()
        if output != '':
            print(output.strip())
        if errout != '':
            print (errout.strip())




if __name__ == '__main__':
    main()
    # gs.gensolution(os.path.realpath(__file__), config, inports, outports)




import io
import json
import os
import re
import subprocess

import sdi_utils.gensolution as gs
import deprecated.set_logging as slog

try:
    api
except NameError:
    class api:
        class Message:
            def __init__(self,body = None,attributes = ""):
                self.body = body
                self.attributes = attributes
                
        def send(port,msg) :
            if isinstance(msg,api.Message) :
                with open(os.path.join('/Users/Shared/data/onlinemedia/prep_texts', msg.attributes['storage.filename']), \
                          mode='w') as f:
                    f.write(json.dumps(msg.body,ensure_ascii=False))
    
        def call(config,jstream):
            api.config = config
            return process(jstream)
            
        def set_port_callback(port, callback) :
            pass
    
        class config:
            ## Meta data
            config_params = dict()
            tags = {'sdi_utils':''}
            version = "0.0.1"
            operator_description = "JSON stream to dict"
            operator_name = 'json_dict'
            operator_description_long = "Converts json stream to dict"
            add_readme = dict()
            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}
            collect = True
            config_params['collect'] = {'title': 'Collect data', 'description': 'Collect data before sending it to the output port',
                                           'type': 'boolean'}

result_jdict = list()


def process(msg) :
    att_dict = msg.attributes

    global result_jdict

    att_dict['operator'] = 'json_dict'
    logger, log_stream = slog.set_logging(att_dict['operator'], loglevel=api.config.debug_mode)
    logger.info("Process started. Logging level: {}".format(logger.level))

    logger.debug('Attributes: {}'.format(str(msg.attributes)))

    if msg.body == None:
        logger.info('Process ended.')
        msg = api.Message(attributes=att_dict, body=result_jdict)
        log = log_stream.getvalue()
        return log, msg
    elif isinstance(msg.body, str):
        json_io = io.StringIO(msg.body)
        logger.debug("Input format: <string>")
    elif isinstance(msg.body, bytes):
        json_io = io.BytesIO(msg.body)
        logger.debug("Input format: <bytes>")
    elif isinstance(msg.body, io.BytesIO):
        logger.debug("Input format: <io.Bytes>")
        json_io = msg.body
    else:
        raise TypeError('Message body has unsupported type' + str(type(msg.body)))

    jdict = json.load(json_io)
    if not isinstance(jdict,list) :
        jdict = [jdict]
    result_jdict.extend(jdict)

    logger.debug('Collect mode: {}'.format(api.config.collect))

    if  (api.config.collect == False) or \
        ('message.lastBatch' in att_dict and att_dict['message.lastBatch'] == True) or \
        ('batch.index' in att_dict and  att_dict['batch.index'] + 1 == att_dict['batch.number']) or \
        ('storage.fileIndex' in att_dict and att_dict['storage.fileIndex'] + 1 == att_dict['storage.fileCount']) :

        msg = api.Message(attributes=att_dict,body=result_jdict)
        api.send(outports[1]['name'], msg)
        logger.info('Msg send to port: {}'.format(outports[1]['name']))
        result_jdict = list()

    api.send(outports[0]['name'], log_stream.getvalue())


inports = [{'name': 'stream', 'type': 'message.file',"description":"Input json byte or string"}]
outports = [{'name': 'log', 'type': 'string',"description":"Logging data"}, \
            {'name': 'data', 'type': 'message.dicts',"description":"Output data as list of dictionaries"}]


#api.set_port_callback(inports[0]['name'], process)

def test_operator() :
    config = api.config
    config.debug_mode = True
    config.collect = False

    in_dir = '/Users/Shared/data/onlinemedia/crawled_texts'
    files_in_dir = [f for f in os.listdir(in_dir) if os.path.isfile(os.path.join(in_dir, f)) and re.match('.*json', f)]
    for i, fname in enumerate(files_in_dir):
        fbase = fname.split('.')[0]
        eos = True if len(files_in_dir) == i + 1 else False
        attributes = {'format': 'csv', "storage.filename": fbase, 'storage.endOfSequence': eos, \
                      'storage.fileIndex': i, 'storage.fileCount': len(files_in_dir),'process_list':[]}
        jsonstream = open(os.path.join(in_dir, fname), mode='rb').read()
        msg = api.Message(attributes=attributes, body=jsonstream)
        api.call(config=config, jstream = msg)


if __name__ == '__main__':
    test_operator()
    if True :
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name+'_'+api.config.version
        subprocess.run(["vctl", "solution", "bundle", '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_0.0.1',\
                                  "-t", solution_name])

        subprocess.run(["mv", solution_name+'.zip', '../../../solution/operators'])
        




import os
import re
import json
import subprocess
import requests
import logging
import argparse
import importlib.util

from shutil import copyfile, move, rmtree

exclude_files = ['__init__.py','setup.py']
exclude_dirs = ['__pycache__','dist','build']

def gensolution(script_path, config, inports, outports, override_readme = False, tar = False, version = None) :

    script_path = os.path.abspath(script_path)
    drive, project_path = os.path.splitdrive(script_path)
    project_path, folder = os.path.split(project_path)
    while folder  and not folder == 'src':
        project_path, folder = os.path.split(project_path)

    src_path = os.path.join(project_path,"src")

    script_filename  = os.path.basename(script_path)

    # generate configSchema.json
    operator_path = os.path.dirname(script_path)
    operator_path_name = operator_path[len(src_path) + 1:]
    package_name = operator_path_name.split(os.path.sep, 1)[0]
    operator_id = operator_path_name.replace(os.path.sep, ".")
    idstr = operator_id + '.configSchema.json'
    configschema = dict()
    configschema["$schema"] =  "http://json-schema.org/draft-06/schema#"
    configschema["$id"] = "http://sap.com/vflow/" + idstr
    configschema["type"] =  "object"
    configschema["properties"] = {'codelanguage' :{'type' :'string'} ,'script' :{'type' :'string'}}
    for k ,v in config.config_params.items() :
        configschema['properties'][k] = v
    logging.info('Write configSchema.json')
    with open(os.path.join(operator_path, 'configSchema.json'), 'w') as schemafile:
        json.dump(configschema, schemafile, indent=4)

    # generate operator.json
    opjson = dict()
    opjson['description'] = config.operator_description
    opjson['component'] = "com.sap.system.python3Operator"
    opjson['inports'] = inports
    opjson['outports'] = outports
    opjson['config'] = {'$type' :configschema["$id"] ,'script' :'file://' +os.path.basename(script_filename)}
    for k ,v in config.config_params.items() :
        opjson['config'][k] = eval('config. ' +k)
    opjson['tags'] = config.tags
    svg_files = [f for f in os.listdir(operator_path) if os.path.isfile(os.path.join(operator_path, f)) and re.match('.*svg', f)]
    if svg_files :
        opjson["iconsrc"] = os.path.basename(svg_files[0])
    logging.info('Write operator.json')
    with open(os.path.join(operator_path, 'operator.json'), 'w') as opfile:
        json.dump(opjson, opfile, indent=4)

    # create README (but only if not exists to prevent overriding manual additions
    readme_file = os.path.join(operator_path, 'README.md')
    if not os.path.exists(readme_file) or override_readme :
        with open(readme_file, 'w') as rmfile:
            logging.info('Write README.md')
            rmfile.write("# {} - {} (Version: {})\n\n".format(config.operator_description,operator_id,config.version))
            if config.operator_description_long :
                rmfile.writelines(config.operator_description_long+'\n\n')
            rmfile.write("## Inport" + '\n\n')
            for ip in opjson['inports'] :
                desc = '' if not 'description' in ip.keys() else ip['description']
                rmfile.write("* **{}** (Type: {}) {}\n".format(ip["name"],ip["type"],desc))
            rmfile.write('\n'+'## outports' + '\n\n')
            for ip in opjson['outports'] :
                desc = '' if not 'description' in ip.keys() else ip['description']
                rmfile.write("* **{}** (Type: {}) {}\n".format(ip["name"],ip["type"],desc))
            rmfile.write('\n'+'## Config' + '\n\n')
            for name,ic in config.config_params.items() :
                rmfile.write("* **{}** - {} (Type: {}) {}\n".format(name,ic["title"],ic["type"],ic["description"]))
            rmfile.write("\n\n")
            if config.tags :
                rmfile.write("# Tags\n")
                for t,v in config.tags.items():
                    rmfile.write("{} : {}".format(t,v))
            rmfile.write("\n\n")
            if config.add_readme :
                for k,s in config.add_readme.items() :
                    rmfile.write("# {}\n".format(k))
                    rmfile.write("{}\n\n".format(s))
        rmfile.close()

        # create __init__.py
        init_name = os.path.join(operator_path, '__init__.py')
        if not os.path.exists(init_name) :
            with open(init_name, 'w') as init_file :
                init_file.close()

    # create dirs and copy data
    if not version :
        version = config.version
    operator_solution_path = os.path.join(project_path, 'solution', 'operators', package_name + '_' + version, 'content', \
                                          'files' ,'vflow' ,'subengines' ,'com' ,'sap', 'python36', 'operators', operator_path_name)
    os.makedirs(operator_solution_path, exist_ok=True)
    logging.info('Copy operator files to: ' + operator_solution_path)
    copyfile(os.path.join(operator_path,"README.md"),os.path.join(operator_solution_path,"README.md"))
    copyfile(os.path.join(operator_path, 'operator.json'), os.path.join(operator_solution_path, 'operator.json'))
    copyfile(os.path.join(operator_path, 'configSchema.json'), os.path.join(operator_solution_path, 'configSchema.json'))
    copyfile(os.path.join(operator_path, '__init__.py'),os.path.join(operator_solution_path, '__init__.py'))

    if svg_files :
        svg_file =  os.path.basename(svg_files[0])
        copyfile(os.path.join(operator_path, svg_file),os.path.join(operator_solution_path, svg_file))

    # Review code and do some adjustments
    # * remove comment in line #api.set_port_callback (used when run locally)
    # * remove all lines following "if __name__ == '__main__': "^
    # * remove import sdi_utils - maybe used in same script for generating
    with open(script_path, 'r') as read_fn:
        with open(os.path.join(operator_solution_path, script_filename), 'w') as write_fn:
            line = read_fn.readline()

            exclude_section = False
            while line :
                if re.match('try:',line) :
                    exclude_section = True
                    line = read_fn.readline()
                    continue

                if exclude_section and re.match('[\w#]+',line) and not re.match('except NameError:',line):
                    exclude_section = False

                # activate call function
                line = re.sub(r'^#\s*api.set_port_callback', 'api.set_port_callback', line)
                line = re.sub(r'^#\s*api.add_generator', 'api.add_generator', line)
                line = re.sub(r'^#\s*api.add_timer', 'api.add_timer', line)
                line = re.sub(r'^#\s*api.add_shutdown_handler', 'api.add_shutdown_handler', line)

                # remove unnecessary imports
                if re.match(r'import subprocess',line) or re.match(r'import sdi_utils.gensolution as gs',line) :
                    line = read_fn.readline()
                    continue

                # end of important part
                if re.match(r'if __name__ == \'__main__\'',line) or re.match(r'if __name__ == \"__main__\"',line) or \
                   re.match(r'def test_operator()',line):
                    break

                write_fn.write(line)
                line = read_fn.readline()

        write_fn.close()
    read_fn.close()

    # create manifest
    manifest = {"name": package_name, "version": version, "format": "2", "dependencies": []}
    root_solution_path = os.path.join(project_path, 'solution', 'operators')
    manifest_filename = os.path.join(root_solution_path, package_name +'_' + version, 'manifest.json')
    logging.info('Write manifest.json ')
    with open(manifest_filename, 'w') as manifestfile:
        json.dump(manifest, manifestfile, indent=4)

def change_version(manifest_file, version) :
    with open(manifest_file,'r') as json_file:
        manifest_dict = json.load(json_file)
        json_file.close()
    manifest_dict['version'] = version
    # versions needs to have a specific format 0.0.0

    with open(manifest_file,'w') as json_file:
        json.dump(manifest_dict, json_file,indent=4)
        json_file.close()

def download_templatefile(url,path) :
    logging.info("Download templateCode.py from Github {} to  {}".format(url,path))
    example_code = requests.get(url)
    open(path, 'wb').write(example_code.content)

###############################################################################################################
### REVERSE
### Load json files configSchema.json and operator.json
### Copy all files to src/operator folder
###############################################################################################################
def make_data_list(port_list) :
    ret_list = list()
    msg_count = 0
    str_count = 0
    data_count = 0
    for ip in port_list:
        if ip['name'] in ['Info','log','info','Log'] and 'string' in ip['type'] :
            ret_list.append('log')
        elif 'message' in ip['type']:
            ret_list.append('msg' + str(msg_count) if msg_count > 0 else 'msg')
            msg_count += 1
        elif 'string' in ip['type']:
            ret_list.append('data_str' + str(str_count) if str_count > 0 else 'data_str')
            str_count += 1
        else:
            ret_list.append('data' + str(data_count) if str_count > 0 else 'data')
            data_count += 1
    return ret_list, ', '.join(ret_list)


def reverse_solution(project_path, package, operator_folder,override=False) :
    operator_folder = operator_folder.replace('.','/')
    operator_solution_path = os.path.join(project_path, 'solution', 'operators', package, 'content','files' ,'vflow', \
                                   'subengines' ,'com' ,'sap', 'python36', 'operators', operator_folder)
    operator_script = os.path.basename(operator_folder) + '.py'
    operator_src_path = os.path.join(project_path, 'src', operator_folder)
    script_src_path = os.path.join(operator_src_path, operator_script)
    if os.path.isfile(script_src_path) :
        if not override :
            raise FileExistsError (script_src_path)
        else :
            logging.warning('Overwrite File: ' + script_src_path)

    found_version = re.match('.+(\d+)\.(\d+)\.(\d+).*',package)
    new_version = "'0.0.1'"
    if found_version and not new_version == '0.0.1' :
        new_version ="'{}.{}.{}'".format(found_version[1],found_version[2],int(found_version[3])+1)


    logging.debug("Load <configSchema.json>")
    with open(os.path.join(operator_solution_path,'configSchema.json')) as schemaFile:
        schema_dict = json.load(schemaFile)
        schemaFile.close()

    logging.debug("Load <operator.json>")
    with open(os.path.join(operator_solution_path,'operator.json')) as operatorFile:
        operator_dict = json.load(operatorFile)
        operatorFile.close()

    logging.debug("Make src-directory for operator")

    logging.debug("Make src-directory for operator: {}".format(operator_src_path))
    os.makedirs(operator_src_path, exist_ok=True)

    logging.debug('Copy files to src-directory for operator (README.md, operator.json, configSchema.json')
    if os.path.isfile(os.path.join(operator_solution_path, "README.md")) :
        copyfile(os.path.join(operator_solution_path, "README.md"), os.path.join(operator_src_path, "README.md"))
    else :
        logging.warning("There is no <README.md> file to copy.")
    copyfile(os.path.join(operator_solution_path, 'operator.json'), os.path.join(operator_src_path, 'operator.json'))
    copyfile(os.path.join(operator_solution_path, 'configSchema.json'),os.path.join(operator_src_path, 'configSchema.json'))
    copyfile(os.path.join(operator_solution_path, operator_script),os.path.join(operator_src_path, 'old_'+operator_script))
    if os.path.isfile(os.path.join(operator_solution_path, operator_dict['iconsrc'])) :
        copyfile(os.path.join(operator_solution_path, operator_dict['iconsrc']), os.path.join(operator_src_path, operator_dict['iconsrc']))
    else :
        logging.warning("There is no icon-file to copy: {}".format(operator_dict['iconsrc']))


    with open(script_src_path,'w') as operatorFile :
        firstblock = r"""import sdi_utils.gensolution as gs
from  sdi_utils import set_logging

try:
    api
except NameError:
    class api:
        class Message:
            def __init__(self,body = None,attributes = ""):
                self.body = body
                self.attributes = attributes
                
        def send(port,msg) :
            if isinstance(msg,api.Message) :
                print('Port: ', port)
                print('Attributes: ', msg.attributes)
                print('Body: ', str(msg.body))
            else :
                print(str(msg))
            return msg
    
        def call(config,msg):
            api.config = config
            return process(msg)
            
        def set_port_callback(port, callback) :
            default_msg = api.Message(attributes={'name':'doit'},body = 'message')
            callback(default_msg)
    
        class config:
            ## Meta data
            config_params = dict()
        """
        operatorFile.write(firstblock)
        operatorFile.writelines("{:4}version = {}\n".format('', new_version))
        if not 'tags' in operator_dict.keys() :
            operatorFile.writelines("{:12}tags = \{\}\n".format(''))
        else :
            operatorFile.writelines("{:12}tags = {}\n".format('', str(operator_dict['tags'])))

        operatorFile.writelines("{:12}operator_description = \"{}\"\n".format('',os.path.basename(operator_folder)))
        for name,defs in schema_dict['properties'].items() :
            if name == 'codelanguage' or name == 'script' :
                continue
            if not name in operator_dict['config'].keys() :
                operator_dict['config'][name] = 'None'
                logging.warning('operator.json has no default value for: {}'.format(name))
            if defs['type'] == 'string' :
                operatorFile.writelines("{:12}{} = '{}'\n".format('', name, operator_dict['config'][name]))
            else :
                operatorFile.writelines("{:12}{} = {}\n".format('',name,operator_dict['config'][name]))
            operatorFile.writelines("{:12}config_params['{}'] = {}\n".format('',name,str(defs)))

        args_list, args_str = make_data_list(operator_dict['inports'])
        ret_list, ret_str = make_data_list(operator_dict['outports'])

        operatorFile.write("\n\ndef process({}) :\n\n".format(args_str))
        operatorFile.write("{:4}logger, log_stream = set_logging('DEBUG')\n\n".format(''))
        operatorFile.write("{:4}# start custom process definition\n".format(''))

        for op in ret_list :
            operatorFile.write("{:4}{} = None\n".format('',op))

        operatorFile.write("{:4}# end custom process definition\n\n".format(''))
        operatorFile.write("{:4}log = log_stream.getvalue()\n".format(''))
        operatorFile.write("{:4}return {}\n\n\n".format('',ret_str))

        operatorFile.write("inports = {}\n".format(str(operator_dict['inports'])))
        operatorFile.write("outports = {}\n\n".format(str(operator_dict['outports'])))

        operatorFile.write("def call_on_input({}) :\n".format(args_str))
        operatorFile.write("{:4}{} = process({})\n".format('',', '.join(ret_list),args_str))

        for i, op in enumerate(operator_dict['outports']) :
            operatorFile.write("{:4}api.send(outports[{}]['name'], {})\n".format('', i, ret_list[i]))

        if len(operator_dict['outports']) > 1 :
            iplist = list()
            for i, ip in enumerate(operator_dict['inports']):
                iplist.append("inports[{}]['name']".format(i))
            setpcallback = "api.set_port_callback([{}], call_on_input)".format(', '.join(iplist))
        else :
            setpcallback = "api.set_port_callback(inports[0]['name'], call_on_input)"
        operatorFile.write('\n#'+setpcallback+'\n\n')

        last2block = r"""def main() :
    print('Test: Default')
    """
        operatorFile.write(last2block)
        operatorFile.write(setpcallback+'\n\n')

        last3block = r"""if __name__ == '__main__':
    main()
    #gs.gensolution(os.path.realpath(__file__), config, inports, outports)
        """
        operatorFile.write(last3block + '\n')
        operatorFile.close()

###############################################################################################################
### MAIN
###############################################################################################################

def main() :

    logging.Formatter('%(levelname)s - %(message)s')

    parser = argparse.ArgumentParser(description='Generate SAP Data Intelligence solution for local operator code')

    parser.add_argument('--project',help='<new project path> - creates new project with required folder structure')
    parser.add_argument('--version', help='<version> in format <num.num.num>')
    parser.add_argument('--debug',action='store_true', help='For debug-level information ')
    parser.add_argument('--zip',action='store_true', help='Zipping solution folder ')
    parser.add_argument('--force', action='store_true', help='Removes subdirectories from <solution/operators>')
    parser.add_argument('--reverse',action='store_true',help='Ceates a custom operator script from solution package')
    parser.add_argument('--package', help='<package name> package in src for solution generation or for reverse custom operator creation')
    parser.add_argument('--operator', help='<package.operator folder> for reverse custom operator creation')

    args = parser.parse_args()
    # testing args
    #args = parser.parse_args(['--project', '../newproject','--force'])
    #args = parser.parse_args(['--project', '.'])
    #args = parser.parse_args(['--version', '0.0.3','--debug','--force'])
    #args = parser.parse_args(['--reverse','--debug','--package','pandasOperators-0.0.16','--operator','sdi_pandas.cleanseHeuristics'])

    version = args.version
    debug = args.debug
    zip_flag = args.zip
    clear_solution_path = args.force
    reverse = args.reverse
    package = args.package
    operator_folder = args.operator


    if debug :
        logging.basicConfig(level=logging.DEBUG,format='%(levelname)s - %(message)s')
        logging.debug('Logging Level: DEBUG')
    else :
        logging.basicConfig(level=logging.INFO,format='%(levelname)s - %(message)s')

    ###############################################################################################################
    ### makes project directories
    ### if --project is dir then the folder name is used as project name
    ### else         the project path is created
    ###############################################################################################################
    if args.project:
        projpathdir = os.path.abspath(os.path.join(args.project, os.pardir))
        if not os.path.isdir(projpathdir):
            logging.error('Path to new project does not exist:  ' + projpathdir)
            exit(-1)

        projpath = args.project
        if not os.path.isabs(projpath):
            projpath = os.path.abspath(projpath)

        projname = os.path.basename(projpath)
        if not os.path.isdir(projpath) :
            logging.info("New project path is created: {}".format(projpath))
            os.mkdir(projpath)

        logging.info('Prepares folder structure at <{}> for project <{}>'.format(projpath,projname))

        src_path = os.path.join(projpath, 'src')
        logging.info('Creates src-directory: ' + src_path)
        os.mkdir(src_path)
        src_project_path = os.path.join(src_path,projname)
        logging.info('Creates package-directory: ' + src_project_path)
        os.mkdir(src_project_path)
        solution_path = os.path.join(projpath, 'solution')
        logging.info('Creates solution-directory: ' + solution_path)
        os.mkdir(solution_path)
        logging.info('Creates README.md')
        readme_file = os.path.join(projpath,'README.md')
        with open(readme_file, 'w') as rmfile:
            readme = '# ' + projname + '\n\nDescription'
            rmfile.write(readme)

        # download templateCode.py
        template_url = 'https://raw.githubusercontent.com/thhapke/gensolution/master/diutil/customOperatorTemplate.py'
        download_templatefile(template_url,os.path.join(src_project_path,"customOperatorTemplate.py"))
        exit(1)

    if version and not re.match(r'\d+\.\d+\.\d+', version) :
        logging.error('<version> does not match required format <d.d.d> :' + version)
        exit(-1)

    ###############################################################################################################
    ### build solution
    ###############################################################################################################
    if not args.reverse :
        if not version :
            logging.error('Option --version required for generating a solution')
            exit(-1)
        project_path = os.getcwd()                              # root path of the whole project
        logging.debug('Current working directory: {}'.format(os.getcwd()))
        src_path = os.path.join(project_path,'src')             # path of the src
        solution_path = os.path.join(project_path, "solution", "operators")

        ### clear solution folder to avoid ambiguities
        if os.path.isdir(solution_path) :
            subdirs = [ f for f in os.listdir(solution_path) if os.path.isdir(f) ]
            if  subdirs :
                logging.warning('Solution path <solution/operators/> not empty')
                if clear_solution_path :
                    logging.info('Remove subdirectory <solution/operators>')
                    rmtree(solution_path, ignore_errors=False, onerror=None)
                else :
                    logging.error('Either clear directory {} manually or run with option --force'.format(solution_path))
                    logging.debug('List of directories in solution/operator folder: ' + str(os.listdir(solution_path)))
                    exit(-1)

        ### generate the json files and copy them to solution directory
        for root, dirs, files in os.walk(os.path.join(src_path,package)):
            for d in dirs:
                if d in exclude_dirs :
                    continue
            for f in files :
                if f in exclude_files :
                    continue
                if re.match(r'.+\.py$',f) :
                    logging.debug('Read specs of {}'.format(os.path.join(root,f)))
                    spec = importlib.util.spec_from_file_location("src", os.path.join(root,f))
                    m = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(m)
                    inports = m.inports if hasattr(m, 'inports') else None
                    gensolution(os.path.join(root,f),config = m.api.config,inports = inports,outports=m.outports,version = version)

        ###  creating operator solutions for uploading
        if zip_flag :
            logging.debug('Zip solution: {}'.format(solution_path))
            for d in os.listdir(solution_path):
                if d in exclude_dirs or re.match(r'.+.zip',d) or re.match(r'^\..+',d) :  # Zips are interpreted as directories
                    logging.debug('Zip skipped: {}'.format(d))
                    continue
                source_dir = os.path.join(solution_path, d)
                logging.debug('Building solution of folder: ' + source_dir)
                if version :
                    logging.debug('Change version in <manifest.json> file')
                    change_version(os.path.join(solution_path,d,'manifest.json'),version = version)
                    d = re.sub(r'(\d+\.\d+\.\d+)$',version,d)
                    dest_dir = os.path.join(solution_path,d)
                    logging.info('Rename folder: {} -> {}'.format(source_dir,dest_dir))
                    move(source_dir, dest_dir)
                    source_dir = os.path.join(solution_path, d)
                tarfilename = os.path.join(solution_path, d + '.zip')
                logging.info('Start vctl cmd: vctl solution bundle {} -t {}'.format(source_dir, tarfilename))
                subprocess.run(["vctl", "solution", "bundle", source_dir, "-t", tarfilename])

    ###############################################################################################################
    ### reverse solution
    ###############################################################################################################
    if args.reverse :
        #project_path = os.path.dirname(os.getcwd()) # for testing only
        project_path = os.getcwd()
        logging.debug('Project path: ' + project_path)
        reverse_solution(project_path,package=package,operator_folder=operator_folder,override=args.force)

if __name__ == '__main__':
    main()



import os

import subprocess

import sdi_utils.gensolution as gs
import deprecated.set_logging as slog

try:
    api
except NameError:
    class api:
        class Message:
            def __init__(self,body = None,attributes = ""):
                self.body = body
                self.attributes = attributes
                
        def send(port,msg) :
            if port == outports[1]['name'] :
                print(msg.attributes)
                print(msg.body)

        class config:
            ## Meta data
            config_params = dict()
            tags = {'sdi_utils':''}
            version = "0.1.0"

            operator_description = "merge attributes"
            operator_name = 'merge_attributes'
            operator_description_long = "Merges the attributes"
            add_readme = dict()
            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}



def process(msg1,msg2) :
    att_dict = dict(msg1.attributes)
    msg1.attributes.update(msg2.attributes)
    msg2.attributes.update(att_dict)

    att_dict['operator'] = 'merge_attributes'
    logger, log_stream = slog.set_logging(att_dict['operator'], loglevel=api.config.debug_mode)

    logger.debug('Attributes: {}'.format(str(msg1.attributes)))

    api.send(outports[0]['name'], log_stream.getvalue())
    api.send(outports[1]['name'], msg1)
    api.send(outports[2]['name'], msg2)


inports = [{'name': 'message1', 'type': 'message',"description":"Input message1"},
           {'name': 'message2', 'type': 'message',"description":"Input message2"}]
outports = [{'name': 'log', 'type': 'string',"description":"Logging data"}, \
            {'name': 'message1', 'type': 'message',"description":"Output message1"},
            {'name': 'message2', 'type': 'message',"description":"Output message2"}]


#api.set_port_callback([inports[0]['name'],inports[1]['name']], process)

def test_operator() :

    api.config.debug_mode = True

    data = [{'name': 'Anna', 'age': 25, 'born': '1995-03-14'}, {'name': 'Berta', 'age': 31, 'born': '1989-08-14'},
            {'name': 'Cecilia', 'age': 41, 'born': '1979-08-14'}]
    msg1 = api.Message(attributes={'format': 'json'}, body=data)
    msg2 = api.Message(attributes={'new':'NEW','more':'MORE'},body='nothing')

    process(msg1,msg2)


if __name__ == '__main__':
    test_operator()
    if True:
        subprocess.run(["rm", '-r',
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version])
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name + '_' + api.config.version
        subprocess.run(["vctl", "solution", "bundle",
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version, \
                        "-t", solution_name])
        subprocess.run(["mv", solution_name + '.zip', '../../../solution/operators'])






import sdi_utils.gensolution as gs
import deprecated.set_logging as slog
import sdi_utils.textfield_parser as tfp
import sdi_utils.tprogress as tp

import pandas as pd
import subprocess
import io
import re
import os

EXAMPLE_ROWS = 5

try:
    api
except NameError:

    class api:

        queue = list()
        class Message:
            def __init__(self, body=None, attributes=""):
                self.body = body
                self.attributes = attributes

        def send(port, msg):
            if port == outports[0]['name']:
                api.queue.append(msg)

        class config:
            ## Meta data
            config_params = dict()
            version = "0.0.17"
            tags = {'pandas': '','sdi_utils':''}
            operator_name = 'csv_df'
            operator_description = "csv to df"
            operator_description_long = "Creating a DataFrame with csv-data passed through inport."
            add_readme = dict()
            add_readme[
                "References"] = "[pandas doc: read_csv](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html)"

            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode', 'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}
            collect = False
            config_params['collect'] = {'title': 'Collect data', 'description': 'Collect data before sending it to the output port',
                                           'type': 'boolean'}
            index_cols = 'None'
            config_params['index_cols'] = {'title': 'Index Columns', 'description': 'Index columns of dataframe',
                                           'type': 'string'}
            separator = ','
            config_params['separator'] = {'title': 'Separator of CSV', 'description': 'Separator of CSV',
                                          'type': 'string'}
            use_columns = 'None'
            config_params['use_columns'] = {'title': 'Use columns from CSV',
                                            'description': 'Use columns from CSV (list)', 'type': 'string'}
            limit_rows = 0
            config_params['limit_rows'] = {'title': 'Limit number of rows',
                                           'description': 'Limit number of rows for testing purpose', 'type': 'number'}
            downcast_int = False
            config_params['downcast_int'] = {'title': 'Downcast integers',
                                             'description': 'Downcast integers from int64 to int with smallest memory footprint',
                                             'type': 'boolean'}
            downcast_float = False
            config_params['downcast_float'] = {'title': 'Downcast float datatypes',
                                               'description': 'Downcast float64 to float32 datatypes',
                                               'type': 'boolean'}

            decimal = '.'
            config_params['decimal'] = {'title': 'Decimals separator', 'description': 'Decimals separator',
                                        'type': 'string'}
            dtypes = 'None'
            config_params['dtypes'] = {'title': 'Data Types of Columns',
                                       'description': 'Data Types of Columns (list of maps)', 'type': 'string'}

            data_from_filename = 'None'
            config_params['data_from_filename'] = {'title': 'Data from Filename', 'description': 'Data from Filename',
                                                   'type': 'string'}
            todatetime = 'None'
            config_params['todatetime'] = {'title': 'To Datetime', 'description': 'To Datetime', 'type': 'string'}
            utc = True
            config_params['utc'] = {'title': 'Use UTC', 'description': 'If true utc is used for time conversion', 'type': 'boolean'}

            keyword_args = "'error_bad_lines'= True, 'low_memory' = False, compression = None, thousands = None "
            config_params['keyword_args'] = {'title': 'Keyword Arguments',
                                             'description': 'Mapping of key-values passed as arguments \"to read_csv\"',
                                             'type': 'string'}
            test_column = 'None'
            config_params['test_column'] = {'title': 'Test columm', 'description': 'Sums up test_column',
                                           'type': 'string'}


def downcast(df, data_type, to_type):
    cols = list(df.select_dtypes(include=[data_type]).columns)
    if len(cols) == 0:
        return df, None

    downcast_dict = dict()
    downcast_dict['data_type'] = data_type
    cdtypes = df[cols].dtypes.to_dict()
    downcast_dict['previous_subtypes'] = {col: str(itype) for col, itype in cdtypes.items()}
    downcast_dict['previous_mem_usage'] = df[cols].memory_usage(deep=True).sum() / 1024 ** 2

    df[cols] = df[cols].apply(pd.to_numeric, downcast=to_type)

    int_dtypes2 = df[cols].dtypes.to_dict()
    downcast_dict['subtypes'] = {col: str(itype) for col, itype in int_dtypes2.items()}
    downcast_dict['mem_usage'] = df[cols].memory_usage(deep=True).sum() / 1024 ** 2

    return df, downcast_dict

result_df = pd.DataFrame()

def process(msg):

    global result_df

    att_dict = msg.attributes

    att_dict['operator'] = 'csv_df'
    logger, log_stream = slog.set_logging(att_dict['operator'], loglevel=api.config.debug_mode)

    logger.info("Process started. Logging level: {}".format(logger.level))
    time_monitor = tp.progress()
    logger.debug('Attributes: {}'.format(str(msg.attributes)))

    if isinstance(msg.body, str):
        csv_io = io.StringIO(msg.body)
        logger.debug("Input format: <string>")
    elif isinstance(msg.body, bytes):
        csv_io = io.BytesIO(msg.body)
        logger.debug("Input format: <bytes>")
    elif isinstance(msg.body, io.BytesIO):
        logger.debug("Input format: <io.Bytes>")
        csv_io = msg.body
    else:
        raise TypeError('Message body has unsupported type' + str(type(msg.body)))

    # nrows
    nrows = None
    if not api.config.limit_rows == 0:
        nrows = api.config.limit_rows

    # usecols
    use_cols = tfp.read_list(api.config.use_columns)
    logger.debug('Columns used: {}'.format(use_cols))

    # dtypes mapping
    typemap = tfp.read_dict(api.config.dtypes)
    logger.debug('Type cast: {}'.format(str(typemap)))

    kwargs = tfp.read_dict(text=api.config.keyword_args, map_sep='=')

    ##### Read string from buffer
    logger.debug("Read from input")
    api.send(outports[0]["name"],log_stream.getvalue())
    log_stream.seek(0)
    log_stream.truncate()
    df = pd.read_csv(csv_io, api.config.separator, usecols=use_cols, dtype=typemap, decimal=api.config.decimal, \
                     nrows=nrows, **kwargs)

    # Data from filename
    if api.config.data_from_filename and not api.config.data_from_filename == 'None':
        col = api.config.data_from_filename.split(':')[0].strip().strip("'").strip('"')
        pat = api.config.data_from_filename.split(':')[1].strip().strip("'").strip('"')
        logger.debug('Filename: {}  pattern: {}'.format(att_dict['filename'], pat))
        try:
            dataff = re.match('.*(\d{4}-\d+-\d+).*', att_dict['filename'])
            df[col] = dataff.group(1)
        except AttributeError:
            raise ValueError('Pattern not found - Filename: {}  pattern: {}'.format(att_dict['filename'], pat))

    # To Datetime
    if api.config.todatetime and not api.config.todatetime == 'None':
        dt_fmt = tfp.read_dict(api.config.todatetime)
        logger.debug('Time conversion {} by using UTC {}'.format(api.config.todatetime,api.config.utc))
        for col, fmt in dt_fmt.items() :
            df[col] = pd.to_datetime(df[col], format=fmt, utc= api.config.utc)

    ###### Downcasting
    # save memory footprint for calculating the savings of the downcast
    logger.debug('Memory used before downcast: {}'.format(df.memory_usage(deep=True).sum() / 1024 ** 2))
    if api.config.downcast_int:
        df, dci = downcast(df, 'int', 'unsigned')
    if api.config.downcast_float:
        df, dcf = downcast(df, 'float', 'float')

    # check if index is provided and set
    index_list = tfp.read_list(api.config.index_cols)
    if index_list:
        df.set_index(index_list, inplace=True)

    if not result_df.empty :
        result_df = pd.concat([result_df, df], axis=0, sort=False)
    else :
        result_df = df

    logger.debug('Columns: {}'.format(str(df.columns)))
    logger.debug('Shape (#rows - #columns): {} - {}'.format(df.shape[0],df.shape[1]))
    logger.debug('Memory: {} kB'.format(df.memory_usage(deep=True).sum() / 1024 ** 2))
    example_rows = EXAMPLE_ROWS if df.shape[0] > EXAMPLE_ROWS else df.shape[0]
    for i in range(0, example_rows):
        logger.debug('Row {}: {}'.format(i,str([str(i)[:10].ljust(10) for i in df.iloc[i, :].tolist()])))

    logger.info('Collecting incoming data: {}'.format(api.config.collect))
    if not api.config.collect or ('message.lastBatch' in msg.attributes and msg.attributes['message.lastBatch'] == True) :
        logger.debug('Send data to outport')
        test_column = tfp.read_value(api.config.test_column)
        if test_column :
            sumup = df[test_column].sum()
            att_dict['Test_Column'] = {'name':test_column,'sum':sumup}
            logger.info('Testcolumn \'{}\': {}'.format(test_column,sumup))
        api.send(outports[1]['name'], api.Message(attributes=att_dict, body=result_df))


    logger.debug('Process ended: {}  '.format(time_monitor.elapsed_time()))
    api.send(outports[0]['name'],log_stream.getvalue())

inports = [{'name': 'csv', 'type': 'message.file',"description":"Input byte or string csv"}]
outports = [{'name': 'log', 'type': 'string',"description":"Logging data"}, \
            {'name': 'data', 'type': 'message.DataFrame',"description":"Output data"}]


#api.set_port_callback(inports[0]['name'], process)

def test_operator():

    api.config.debug_mode = True
    api.config.use_columns = "'Exportdatum','Postleitzahl','Ort','Ortsteil','Verbrauchsstufe','Rang','Gesamtpreis','Anbietername'"
    api.config.downcast_float = True
    api.config.downcast_int = True
    api.config.dtypes = "'Gesamtpreis':'float32','Postleitzahl':'uint32','Verbrauchsstufe':'uint16'"
    api.config.separator = ';'
    api.config.index_cols = "None"
    api.config.limit_rows = 0
    api.config.decimal = '.'
    api.config.utc = False
    api.config.collect = False
    api.config.todatetime = 'Exportdatum : %Y-%m-%d'
    api.config.test_column = 'Postleitzahl'
    api.config.keyword_args = "'error_bad_lines'= True, 'low_memory' = False, compression = None, comment = '#'"


    in_dir = '/Users/Shared/data/OptRanking/portal1_samples25'
    files_in_dir = [f for f in os.listdir(in_dir) if os.path.isfile(os.path.join(in_dir, f)) and re.match('.*csv', f)]

    for i, fname in enumerate(files_in_dir):

        if i == 5 :
            break

        fbase = fname.split('.')[0]
        eos = True if len(files_in_dir) == i + 1 else False
        attributes = {'format': 'csv', "storage.filename": fbase, 'storage.endOfSequence': eos, \
                      'storage.fileIndex': i, 'storage.fileCount': len(files_in_dir),'message.lastBatch':eos}
        csv = open(os.path.join(in_dir, fname), mode='rb').read()
        msg = api.Message(attributes=attributes, body=csv)

        process(msg)

if __name__ == '__main__':
    #test_operator()

    if True :
        subprocess.run(["rm", '-r',
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version])
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name + '_' + api.config.version
        subprocess.run(["vctl", "solution", "bundle",
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version, \
                        "-t", solution_name])
        subprocess.run(["mv", solution_name + '.zip', '../../../solution/operators'])




import time
import subprocess
import os

import sdi_utils.gensolution as gs
import deprecated.set_logging as slog
import sdi_utils.textfield_parser as tfp

try:
    api
except NameError:
    class api:
        class Message:
            def __init__(self, body=None, attributes=""):
                self.body = body
                self.attributes = attributes

        def send(port, msg):
            if isinstance(msg, api.Message):
                print('{}: {}'.format(port, msg.body))
            else:
                print('{}: {}'.format(port, msg))

        class config:
            ## Meta data
            config_params = dict()
            tags = {'pandas': '', 'sdi_utils': ''}
            version = "0.1.0"
            operator_name = 'gate'
            operator_description = "Gate"
            operator_description_long = "Gate that counts incoming messages and sends out message with latest \
message when gate limit has been reached."
            add_readme = dict()

            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}
            limit = 100000000
            config_params['limit'] = {'title': 'Limit', 'description': \
                'Limit after which the message is send', 'type': 'integer'}
            sleep = 0
            config_params['sleep'] = {'title': 'Sleep', 'description': \
                'Time before starting next processing step', 'type': 'integer'}

            attribute = None
            config_params['attribute'] = {'title': 'Attribute', 'description': \
                'Attribute that keeps a number used as limit, e.g. storage.fileCount', 'type': 'string'}


inports = [{"name": "limit", "type": "int64", "description": "Limit"}, \
            {"name": "attributes", "type": "message", "description": "Message with attributes to used"},\
           {"name": "data", "type": "message.*", "description": "Input data"}]
outports = [{'name': 'log', 'type': 'string', "description": "Logging data"}, \
            {'name': 'trigger', 'type': 'string', "description": "Trigger next process step"},
            {"name": "data", "type": "message", "description": "Output of unchanged last input data"}]

counter = 0
logger, log_stream = slog.set_logging('Gate', loglevel=api.config.debug_mode)

def call_on_message(msg):
    global counter
    logger, log_stream = slog.set_logging('Gate', loglevel=api.config.debug_mode)
    counter += 1

    if 'message.lastBatch' in msg.attributes and msg.attributes['message.lastBatch'] == True:
        time.sleep(api.config.sleep)
        api.send(outports[2]['name'], msg)
        api.send(outports[1]['name'], '1')
        logger.info('Msg send to outports: Last batch in attributes')
    elif counter >= api.config.limit:
        time.sleep(api.config.sleep)
        api.send(outports[2]['name'], msg)
        api.send(outports[1]['name'], '1')
        logger.info(
            'Msg send to outports: Counter greater or equal than limit: {} >= {}'.format(counter, api.config.limit))
    else:
        logger.info('No message send')

    logger.debug('Counter: {} / {}'.format(counter, api.config.limit))
    logger.debug('Message Attributes: {}'.format(msg.attributes))
    api.send(outports[0]['name'], log_stream.getvalue())



def call_on_limit(new_limit):
    logger, log_stream = slog.set_logging('Gate', loglevel=api.config.debug_mode)
    old_limit = api.config.limit
    api.config.limit = new_limit
    logger.info('Set new limit to: {} -> {}'.format(old_limit, api.config.limit))
    api.send(outports[0]['name'], log_stream.getvalue())


def call_on_attribute(msg):
    logger, log_stream = slog.set_logging('Gate', loglevel=api.config.debug_mode)
    old_limit = api.config.limit
    limit_attribute = tfp.read_value(api.config.attribute)
    if not limit_attribute :
        raise ValueError('Attribute has not been set in gate parameters.')
    api.config.limit = msg.attributes[limit_attribute]
    logger.info('Set new limit to: {} -> {}'.format(old_limit, api.config.limit))
    api.send(outports[0]['name'], log_stream.getvalue())




def call_on_limit(new_limit):
    logger, log_stream = slog.set_logging('Gate', loglevel=api.config.debug_mode)
    old_limit = api.config.limit
    api.config.limit = new_limit
    logger.info('Set new limit to: {} -> {}'.format(old_limit, api.config.limit))
    api.send(outports[0]['name'], log_stream.getvalue())


def call_on_attribute(msg):
    logger, log_stream = slog.set_logging('Gate', loglevel=api.config.debug_mode)
    old_limit = api.config.limit
    limit_attribute = tfp.read_value(api.config.attribute)
    if not limit_attribute :
        raise ValueError('Attribute has not been set in gate parameters.')
    api.config.limit = msg.attributes[limit_attribute]
    logger.info('Set new limit to: {} -> {}'.format(old_limit, api.config.limit))
    api.send(outports[0]['name'], log_stream.getvalue())


#api.set_port_callback(inports[0]['name'], call_on_limit)
#api.set_port_callback(inports[1]['name'], call_on_attribute)
#api.set_port_callback(inports[2]['name'], call_on_message)



def test_operator():
    api.config.sleep = 0
    api.config.limit = 5
    api.config.attribute = 'storage.fileCount'

    for i in range(0, 0):
        msg = api.Message(attributes={'msg': 'test'}, body=i)
        call_on_message(msg)
        if i == 6:
            call_on_limit(9)

    trigger_msg = api.Message(attributes={'storage.fileCount': 3},body='msgbody')
    call_on_attribute(trigger_msg)
    for i in range(0, 10):
        msg = api.Message(attributes={'msg': 'test'}, body=i)
        call_on_message(msg)


if __name__ == '__main__':
    #test_operator()
    if True :
        subprocess.run(["rm", '-r',
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators' + api.config.version])
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name + '_' + api.config.version
        subprocess.run(["vctl", "solution", "bundle",
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version, \
                        "-t", solution_name])
        subprocess.run(["mv", solution_name + '.zip', '../../../solution/operators'])


import subprocess
import os
import pandas as pd

import sdi_utils.gensolution as gs
import deprecated.set_logging as slog
import sdi_utils.textfield_parser as tfp
import sdi_utils.tprogress as tp

try:
    api
except NameError:
    class api:
        class Message:
            def __init__(self,body = None,attributes = ""):
                self.body = body
                self.attributes = attributes

        def send(port,msg) :
            if port == outports[1]['name'] :
                print('ATTRIBUTES: ')
                print(msg.attributes)#
                print('CSV-String: ')
                print(msg.body)

        class config:
            ## Meta data
            config_params = dict()
            tags = {'sdi_utils':''}
            version = "0.0.1"
            operator_name = 'table_csv'
            operator_description = "table to csv"
            operator_description_long = "Converts table to csv stream."
            add_readme = dict()
            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}

            drop_columns = 'None'
            config_params['drop_columns'] = {'title': 'Drop Columns',
                                           'description': 'List of columns to drop.',
                                           'type': 'string'}


def process(msg):

    att = dict(msg.attributes)
    att['operator'] = 'table_csv'
    logger, log_stream = slog.set_logging(att['operator'], loglevel=api.config.debug_mode)

    logger.info("Process started")
    time_monitor = tp.progress()

    header = [c["name"] for c in msg.attributes['table']['columns']]
    df = pd.DataFrame(msg.body,columns=header)

    drop_columns = tfp.read_list(api.config.drop_columns)
    if drop_columns :
        df = df.drop(columns = drop_columns)

    csv = df.to_csv(index=False)
    att["file"] =  {"connection": {"configurationType": "Connection Management", "connectionID": "unspecified"}, \
                    "path": "open", "size": 0}

    msg = api.Message(attributes=att,body = csv)
    api.send(outports[1]['name'],msg)

    api.send(outports[0]['name'], log_stream.getvalue())


inports = [{'name': 'data', 'type': 'message.table',"description":"Input message with table"}]
outports = [{'name': 'log', 'type': 'string',"description":"Logging data"}, \
            {'name': 'csv', 'type': 'message.file',"description":"Output data as csv"}]


#api.set_port_callback(inports[0]['name'], process)

def test_operator() :

    api.config.debug_mode = True
    api.config.drop_columns = "header2"

    attributes = {"table":{"columns":[{"class":"string","name":"header1","nullable":True,"size":80,"type":{"hana":"NVARCHAR"}},
                                      {"class":"string","name":"header2","nullable":True,"size":3,"type":{"hana":"NVARCHAR"}},
                                      {"class":"string","name":"header3","nullable":True,"size":10,"type":{"hana":"NVARCHAR"}}],
                           "name":"test.table","version":1}}
    table = [ [(j*3 + i) for i in range(0,3)] for j in range (0,5)]
    msg = api.Message(attributes=attributes, body=table)
    print(table)
    process(msg)


if __name__ == '__main__':
    #test_operator()
    if True :
        print(os.getcwd())
        subprocess.run(["rm", '-r','../../../solution/operators/sdi_utils_operators_' + api.config.version])
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name + '_' + api.config.version
        subprocess.run(["vctl", "solution", "bundle",'../../../solution/operators/sdi_utils_operators_' + api.config.version, \
                        "-t", solution_name])
        subprocess.run(["mv", solution_name + '.zip', '../../../solution/operators'])



import sdi_utils.gensolution as gs
import deprecated.set_logging as slog
import sdi_utils.tprogress as tp

import subprocess
import os
from datetime import datetime, timezone
import pandas as pd
import random

pd.set_option('mode.chained_assignment',None)

try:
    api
except NameError:
    class api:

        queue = list()

        class Message:
            def __init__(self, body=None, attributes=""):
                self.body = body
                self.attributes = attributes

        def send(port, msg):
            if port == outports[1]['name']:
                api.queue.append(msg)

        class config:
            ## Meta data
            config_params = dict()
            version = '0.0.1'
            tags = {'pandas': '', 'sdi_utils': ''}
            operator_name = 'df_replication_cols'
            operator_description = "Adding replciation columns"

            operator_description_long = "Adds replications columns to a table."
            add_readme = dict()
            add_readme["References"] = ""

            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}

            package_size = 1000
            config_params['package_size'] = {'title': 'Package size',
                                       'description': 'Package size',
                                       'type': 'integer'}


def process(msg):
    att_dict = msg.attributes

    att_dict['operator'] = 'df_replication_cols'
    logger, log_stream = slog.set_logging(att_dict['operator'], loglevel=api.config.debug_mode)

    logger.info("Process started. Logging level: {}".format(logger.level))
    time_monitor = tp.progress()
    logger.debug('Attributes: {}'.format(str(msg.attributes)))

    df = msg.body
    logger.debug('DataFrame: {} - {}'.format(df.shape[0],df.shape[1]))

    #df['DATE'] = pd.to_datetime(df['DATE'], format='%Y-%m-%d', utc=False)

    df['status'] = 'W'
    dt = datetime.now(timezone.utc)
    packageid = dt.strftime("%Y%m%d%H%M%S")
    df['updated'] = dt.strftime("%Y-%m-%d %H:%M:%S")
    #df['packageid'] = df['packageid'].astype('int64')
    df['packageid'] = 0
    att_dict['packageid'] = packageid
    att_dict['updated'] = dt

    # packageid creation
    packageid_start = int(random.random() * 10000)
    package_size = api.config.package_size
    for i, start in enumerate(range(0, df.shape[0], package_size)):
        df.packageid.iloc[start:start + package_size] = packageid_start + i


    logger.debug('Process ended: {}'.format(time_monitor.elapsed_time()))
    api.send(outports[0]['name'], log_stream.getvalue())
    api.send(outports[1]['name'], api.Message(attributes=att_dict, body=df))


inports = [{'name': 'table', 'type': 'message.DataFrame', "description": "DataFrame"}]
outports = [{'name': 'log', 'type': 'string', "description": "Logging data"}, \
            {'name': 'table', 'type': 'message.DataFrame', "description": "DataFrame with replication columns"}]


# api.set_port_callback(inports[0]['name'], process)

def test_operator():
    api.config.package_size = 1
    df = pd.DataFrame(
        {'icol': [1, 2, 3, 4, 5], 'DATE': ['2020-01-01', '2020-02-01', '2020-01-31', '2020-01-28', '2020-04-12'], \
         'col3': [100.0, 200.2, 300.4, 400, 500], 'names': ['Anna', 'Berta', 'Berta', 'Claire', 'Dora'], \
         'bool': [True, False, False, True, True]})

    attributes = {'format': 'csv', 'name': 'DF_name', 'process_list': []}
    msg = api.Message(attributes=attributes, body=df)
    process(msg)

    df_list = [d.body for d in api.queue]
    df = pd.concat(df_list)
    print(df)
    print(df.dtypes)


if __name__ == '__main__':
    #test_operator()
    if True:
        subprocess.run(["rm", '-r',
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version])
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name + '_' + api.config.version
        subprocess.run(["vctl", "solution", "bundle",
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version, \
                        "-t", solution_name])
        subprocess.run(["mv", solution_name + '.zip', '../../../solution/operators'])




import io
import subprocess
import os
import pandas as pd

import sdi_utils.gensolution as gs
import deprecated.set_logging as slog
import sdi_utils.textfield_parser as tfp

try:
    api
except NameError:
    class api:
        class Message:
            def __init__(self, body=None, attributes=""):
                self.body = body
                self.attributes = attributes

        def send(port, msg):
            if port == outports[1]['name']:
                print(msg.attributes)
                print(msg.body)

        class config:
            ## Meta data
            config_params = dict()
            tags = {'sdi_utils': '', 'pandas':'','numpy36':''}
            version = "0.0.1"
            operator_description = "csv to table create"
            operator_name = 'csv_table_create'
            operator_description_long = "csv to table create"
            add_readme = dict()
            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}

            table_name = 'table_name'
            config_params['table_name'] = {'title': 'Table Name',
                                           'description': 'Table name.',
                                           'type': 'string'}



def process(msg):
    att = dict(msg.attributes)

    att['operator'] = 'csv_table_create'
    logger, log_stream = slog.set_logging(att['operator'], loglevel=api.config.debug_mode)

    logger.debug('Attributes: {}'.format(str(att)))

    csv_io = io.BytesIO(msg.body)

    df = pd.read_csv(csv_io)

    table_name = api.config.table_name
    if not tfp.read_value(table_name) :
        table_name = os.path.basename(att['file']['path']).split('.')[0]

    att["table"] = {"name": table_name, "version": 1, "columns": list()}

    for col in df.columns :
        col_dtype = str(df[col].dtype)
        if "int" in col_dtype:
            att["table"]["columns"].append({"name": col, "type":{"hana":"BIGINT"}})
        elif "float" in  col_dtype :
            att["table"]["columns"].append({"name": col, "type": {"hana": "DOUBLE"}})
        elif 'object' in col_dtype  :
            max = df[col].str.len().max()
            att["table"]["columns"].append({"name": col, "size": max + 1,"type": {"hana": "NVARCHAR"}})
        else:
            raise ValueError('Unknown data type: {}'.format(col_dtype))

    sql = "CREATE COLUMN TABLE {table} (".format(table=table_name)
    for col in att['table']['columns'] :
        if col['type']['hana'] == 'NVARCHAR' :
            sql += "\n\t{col_name} {col_type} ({size}),".format(col_name=col['name'], col_type=col['type']['hana'],size =col['size'])
        else :
            sql += "\n\t{col_name} {col_type},".format(col_name = col['name'],col_type = col['type']['hana'])
    sql = sql[:-1] + "\n)"

    api.send(outports[1]['name'], api.Message(attributes=att,body=sql))
    api.send(outports[0]['name'], log_stream.getvalue())


inports = [{'name': 'stream', 'type': 'message.file', "description": "Input csv byte or string"}]
outports = [{'name': 'log', 'type': 'string', "description": "Logging data"}, \
            {'name': 'sql', 'type': 'message.table', "description": "Output data as table"}]

#api.set_port_callback(inports[0]['name'], process)


def test_operator():

    csv = b'id,name,height,birthday\n0,"Albert",1.75,"1960-01-31"\n1,"Berta",1.68,"1965-03-22"\n2,"Claire",1.65,"1970-05-11"'

    csv = pd.read_csv('/Users/d051079/Downloads/car_sales.csv').to_csv(index=False).encode('utf-8')


    attributes = {'format': 'csv', "file": {'path': "/usr/data/friends.csv"}, 'message.lastBatch': True, \
                  'storage.fileIndex': 0, 'storage.fileCount': 1, 'process_list': []}
    msg = api.Message(attributes=attributes, body=csv)
    process(msg)


if __name__ == '__main__':
    test_operator()
    if True :
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name+'_'+api.config.version
        print('Solution name: {}'.format(solution_name))
        print('Current directory: {}'.format(os.getcwd()))
        subprocess.run(["vctl", "solution", "bundle", '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_0.0.1',\
                                  "-t", solution_name])
        subprocess.run(["mv", solution_name+'.zip', '../../../solution/operators'])



import io
import os
import subprocess

import sdi_utils.gensolution as gs
import deprecated.set_logging as slog
import sdi_utils.tprogress as tp

try:
    api
except NameError:
    class api:
        class Message:
            def __init__(self,body = None,attributes = ""):
                self.body = body
                self.attributes = attributes
                
        def send(port,msg) :
            if port == outports[1]['name'] :
                print(msg.attributes)
                for i,dk in enumerate(msg.body) :
                    print(dk)
                    if i > 100 :
                        break
                print(msg.body)
            else :
                print('{}: {}'.format(port,msg))
    
        def call(config,csvstream):
            api.config = config
            return process(csvstream)
            
        def set_port_callback(port, callback) :
            pass
    
        class config:
            ## Meta data
            config_params = dict()
            tags = {'sdi_utils':''}
            version = "0.0.1"
            operator_name = 'line_array'
            operator_description = "Line to array"
            operator_description_long = "Converts lines to array"
            add_readme = dict()
            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}
            collect = True
            config_params['collect'] = {'title': 'Collect data', 'description': 'Collect data before sending it to the output port',
                                           'type': 'boolean'}

            lexicographical = False
            config_params['lexicographical'] = {'title': 'Lexicographical order', 'description': 'Order array according to lexicographical order ',
                                           'type': 'boolean'}


all_lines = list()

def process(msg):
    att_dict = msg.attributes

    global all_lines

    att_dict['operator'] = 'line_array'
    logger, log_stream = slog.set_logging(att_dict['operator'], loglevel=api.config.debug_mode)
    logger.info("Process started")
    time_monitor = tp.progress()

    logger.debug('Attributes: {}'.format(str(msg.attributes)))

    lines_io = io.StringIO(msg.body.decode('utf-8'))
    lines = lines_io.read().splitlines()
    lines = list(filter(None, lines))

    all_lines.extend(lines)

    if api.config.lexicographical :
        all_lines.sort()

    logger.debug('Process ended: {}'.format(time_monitor.elapsed_time()))
    if 'message.lastBatch' in msg.attributes :
        if msg.attributes['message.lastBatch'] or api.config.collect == False:
            msg = api.Message(attributes=att_dict, body=all_lines)
            api.send(outports[1]['name'], msg)
    else :
        msg = api.Message(attributes=att_dict, body=all_lines)
        api.send(outports[1]['name'], msg)

    if api.config.debug_mode:
        example_lines = 100
        logger.debug('Example lines #{}'.format(example_lines))
        for i, dk in enumerate(all_lines):
            logger.debug(dk)
            if i > example_lines:
                break

    api.send(outports[0]['name'], log_stream.getvalue())


inports = [{'name': 'stream', 'type': 'message.file',"description":"Input csv byte or string"}]
outports = [{'name': 'log', 'type': 'string',"description":"Logging data"}, \
            {'name': 'data', 'type': 'message.list',"description":"Output data as list"}]


#api.set_port_callback(inports[0]['name'], process)

def test_operator() :
    config = api.config
    config.debug_mode = True
    config.collect = False
    config.lexicographical = True

    fname = '/Users/Shared/data/onlinemedia/repository/blacklist.txt'

    fbase = fname.split('.')[0]
    attributes = {'format': 'csv', "storage.filename": fbase, 'message.lastBatch': True, \
                  'storage.fileIndex': 0, 'storage.fileCount': 1,'process_list':[]}
    csvstream = open(fname, mode='rb').read()
    msg = api.Message(attributes=attributes, body=csvstream)
    api.call(config=config, csvstream = msg)


if __name__ == '__main__':
    #test_operator()
    if True :
        subprocess.run(["rm", '-r',
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators' + api.config.version])
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name + '_' + api.config.version
        subprocess.run(["vctl", "solution", "bundle",
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version, \
                        "-t", solution_name])
        subprocess.run(["mv", solution_name + '.zip', '../../../solution/operators'])



import subprocess
import os

import sdi_utils.gensolution as gs
import deprecated.set_logging as slog

try:
    api
except NameError:
    class api:
        class Message:
            def __init__(self, body=None, attributes=""):
                self.body = body
                self.attributes = attributes

        def send(port, msg):
            if port == outports[1]['name']:
                print('Message passed: {} - {}'.format(msg.attributes,msg.body))
            elif port == outports[2]['name']:
                print('Message did not pass: {} - {}'.format(msg.attributes,msg.body))

        class config:
            ## Meta data
            config_params = dict()
            tags = {'sdi_utils': ''}
            version = "0.1.0"
            operator_name = 'decision'
            operator_description = "Decision"
            operator_description_long = "Decision gate that channels messages."
            add_readme = dict()

            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}

            decision_attribute = 'message.lastBatch'
            config_params['decision_attribute'] = {'title': 'Descision Attribute',
                                           'description': 'Decision Attribute',
                                           'type': 'string'}


inports = [{"name": "input", "type": "message.*", "description": "Input data"}]
outports = [{'name': 'log', 'type': 'string', "description": "Logging data"}, \
            {'name': 'True', 'type': 'message.*', "description": "True message"},
            {"name": "False", "type": "message.*", "description": "False message"}]


logger, log_stream = slog.set_logging('Gate', loglevel=api.config.debug_mode)

def process(msg):

    logger, log_stream = slog.set_logging('Gate', loglevel=api.config.debug_mode)

    if api.config.decision_attribute in msg.attributes and msg.attributes[api.config.decision_attribute] == True:
        api.send(outports[1]['name'], msg)
        logger.info('Msg passed: {}'.format(msg.attributes))
        api.send(outports[0]['name'], log_stream.getvalue())
    else :
        api.send(outports[2]['name'], msg)
        logger.info('Msg did not pass: {}'.format(msg.attributes))
        api.send(outports[0]['name'], log_stream.getvalue())


#api.set_port_callback(inports[0]['name'], process)


def test_operator():
    #api.config.last_attribute = 'message.last_update'
    process(api.Message(attributes={'message.lastBatch':False},body='0'))
    process(api.Message(attributes={'message.lastBatch':True},body='1'))
    process(api.Message(attributes={'message.last_update': True}, body='2'))


if __name__ == '__main__':
    test_operator()
    if True :
        subprocess.run(["rm", '-r',
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators' + api.config.version])
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name + '_' + api.config.version
        subprocess.run(["vctl", "solution", "bundle",
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version, \
                        "-t", solution_name])
        subprocess.run(["mv", solution_name + '.zip', '../../../solution/operators'])


import json
import os
import re
import subprocess

import pandas as pd

import sdi_utils.gensolution as gs
import deprecated.set_logging as slog

try:
    api
except NameError:
    class api:

        queue = list()

        class Message:
            def __init__(self, body=None, attributes=""):
                self.body = body
                self.attributes = attributes

        def send(port, msg):
            if port == outports[1]["name"]:
                api.queue.append(msg)

        def set_config(config):
            api.config = config

        class config:
            ## Meta data
            config_params = dict()
            tags = {'sdi_utils': ''}
            version = "0.0.1"
            operator_description = "json to df"
            operator_name = 'json_df'
            operator_description_long = "Converts json stream to DataFrame"
            add_readme = dict()
            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}


def process(msg):
    att_dict = msg.attributes

    att_dict['operator'] = 'json_df'
    logger, log_stream = slog.set_logging(att_dict['operator'], loglevel=api.config.debug_mode)
    logger.info("Process started. Logging level: {}".format(logger.level))

    logger.debug('Attributes: {}'.format(str(msg.attributes)))

    jdict = json.loads(msg.body)
    if not isinstance(jdict, list):
        jdict = [jdict]

    df = pd.DataFrame(jdict)
    logger.debug('DataFrame created: {} - {}'.format(df.shape[0], df.shape[1]))

    msg = api.Message(attributes=att_dict, body=df)
    api.send(outports[1]['name'], msg)
    logger.info('Msg send to port: {}'.format(outports[1]['name']))
    api.send(outports[0]['name'], log_stream.getvalue())


inports = [{'name': 'stream', 'type': 'message.file', "description": "Input json byte or string"}]
outports = [{'name': 'log', 'type': 'string', "description": "Logging data"}, \
            {'name': 'data', 'type': 'message.DataFrame', "description": "Output DataFrame"}]


#api.set_port_callback(inports[0]['name'], process)

def test_operator() :

    api.config.debug_mode = True
    api.config.collect = False

    in_dir = '/Users/Shared/data/onlinemedia/crawled_texts'
    files_in_dir = [f for f in os.listdir(in_dir) if os.path.isfile(os.path.join(in_dir, f)) and re.match('.*json', f)]
    for i, fname in enumerate(files_in_dir):
        fbase = fname.split('.')[0]
        eos = True if len(files_in_dir) == i + 1 else False
        attributes = {'format': 'csv', "storage.filename": fbase, 'storage.endOfSequence': eos, \
                      'storage.fileIndex': i, 'storage.fileCount': len(files_in_dir),'process_list':[]}
        jsonstream = open(os.path.join(in_dir, fname), mode='rb').read()
        msg = api.Message(attributes=attributes, body=jsonstream)
        process(msg)

    out_file = '/Users/Shared/data/test/json_df.csv'
    df_list = [d.body for d in api.queue]
    pd.concat(df_list).to_csv(out_file,index=False)


if __name__ == '__main__':

    #test_operator()
    if True :
        subprocess.run(["rm", '-r',
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version])
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name + '_' + api.config.version
        subprocess.run(["vctl", "solution", "bundle",
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version, \
                        "-t", solution_name])
        subprocess.run(["mv", solution_name + '.zip', '../../../solution/operators'])






import pandas as pd
import subprocess
import io
import os

try:
    api
except NameError:

    class api:

        queue = list()
        class Message:
            def __init__(self, body=None, attributes=""):
                self.body = body
                self.attributes = attributes

        def send(port, msg):
            if port == outports[0]['name']:
                api.queue.append(msg)

        class config:
            ## Meta data
            config_params = dict()
            version = "0.0.17"
            tags = {'pandas': ''}
            operator_name = 'pandas_script'
            operator_description = "Pandas Script"
            operator_description_long = "Creating a DataFrame with csv-data passed through inport."
            add_readme = dict()
            add_readme[
                "References"] = "[pandas doc: read_csv](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html)"


def process(msg):

    attributes = dict(msg.attributes)
    csv_io = io.BytesIO(msg.body)
    df = pd.read_csv(csv_io)
    data_str = df.to_csv(sep=',', index=False)
    msg = api.Message(attributes=attributes,body=data_str)
    api.send(outports[0]['name'],msg)


inports = [{'name': 'csv', 'type': 'message.file',"description":"Input csv (byte)"}]
outports = [{'name': 'csv', 'type': 'message.file',"description":"Output csv (string)"}]


#api.set_port_callback(inports[0]['name'], process)

def test_operator():

    fname = '/Users/Shared/data/TechEd2020/input_product_ranks/product_ranks.csv'
    attributes = {'format': 'csv', "storage.filename": fname}
    csv = open(fname, mode='rb').read()
    msg = api.Message(attributes=attributes, body=csv)

    process(msg)

    for q in api.queue:
        print(msg.body)

if __name__ == '__main__':
    test_operator()

    if True :
        subprocess.run(["rm", '-r',
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version])
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name + '_' + api.config.version
        subprocess.run(["vctl", "solution", "bundle",
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version, \
                        "-t", solution_name])
        subprocess.run(["mv", solution_name + '.zip', '../../../solution/operators'])




import time
from datetime import datetime
import re
import subprocess
import os


import sdi_utils.gensolution as gs
import deprecated.set_logging as slog
import sdi_utils.textfield_parser as tfp
import sdi_utils.tprogress as tp

try:
    api
except NameError:
    class api:
        class Message:
            def __init__(self,body = None,attributes = ""):
                self.body = body
                self.attributes = attributes
                
        def send(port,msg) :

            if port == outports[2]['name'] :
                print('Attributes: {}'.format(msg.attributes))
                print('Filename: {}'.format(msg.body))
            elif port == outports[1]['name'] :
                print('Files processed: {}'.format(msg))

        class config:
            ## Meta data
            config_params = dict()
            tags = {'sdi_utils':''}
            version = "0.1.0"
            operator_name = 'filter_date'
            operator_description = "Filter date string"
            operator_description_long = "Filters filenames with date pattern (yyyy-mm-dd or yyyymmdd)."
            add_readme = dict()
            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}
            start_date = 'None'
            config_params['start_date'] = {'title': 'Start Date',
                                           'description': 'Start date. If value is none then today date is taken. Format:YYYY-MM-DD',
                                           'type': 'string'}
            end_date = 'None'
            config_params['end_date'] = {'title': 'End Date',
                                           'description': 'End date. If value is none then today date is taken. Format:YYYY-MM-DD',
                                           'type': 'string'}



file_msg_list = list()

def process(msg):
    global file_msg_list

    att_dict = msg.attributes

    att_dict['operator'] = 'filter_date'
    logger, log_stream = slog.set_logging(att_dict['operator'], loglevel=api.config.debug_mode)
    logger.info("Process started")
    time_monitor = tp.progress()
    today_str = time.strftime("%Y-%m-%d")
    logger.debug("Today string: {}".format(today_str))

    start_date_str = tfp.read_value(api.config.start_date)
    if start_date_str == None:
        start_date_str = today_str
    start_date = datetime.strptime(start_date_str, "%Y-%m-%d")

    end_date_str = tfp.read_value(api.config.end_date)

    if end_date_str == None:
        end_date_str = today_str
    end_date = datetime.strptime(end_date_str, "%Y-%m-%d")

    filename = msg.attributes['file']['path']
    logger.debug('Attributes: {}'.format(str(msg.attributes)))
    logger.debug('Body: {}'.format(str(msg.body)))

    file_date = None
    if re.search('\d{4}-\d{2}-\d{2}', filename):
        file_date_str = re.search('\d{4}-\d{2}-\d{2}', filename).group(0)
        file_date = datetime.strptime(file_date_str, "%Y-%m-%d")
        logger.debug('Date found in filename: {}'.format(file_date.strftime("%Y-%m-%d")))
    elif re.search('(\d{4})(\d{2})(\d{2})', filename):
        dates = re.search('(\d{4})(\d{2})(\d{2})', filename)
        try:
            file_date = datetime(year=int(dates.group(1)), month=int(dates.group(2)), day=int(dates.group(3)))
            logger.debug('Date found in filename: {}'.format(file_date.strftime("%Y-%m-%d")))
        except ValueError:
            file_date = None
    else:
        logger.debug('No date pattern found in filename: {}'.format(filename))

    # Save file msg to list
    if file_date and file_date >= start_date and file_date <= end_date:
        file_msg_list.append(msg)
        logger.info('Date in filename is within given range: {} ({} - {})'.format(file_date.strftime("%Y-%m-%d"),
                                                                                  start_date_str, end_date_str))

    # If lastBatch then send all filenames
    if msg.attributes['message.lastBatch']:
        api.send(outports[1]['name'], len(file_msg_list))
        logger.info('#Filenames passed filter: {}'.format(len(file_msg_list)))
        for i, m in enumerate(file_msg_list):
            m.attributes['message.batchIndex'] = i
            m.attributes['message.batchSize'] = len(file_msg_list)
            m.attributes['message.lastBatch'] = True if i == len(file_msg_list) - 1 else False
            logger.info('File send to outport: {}'.format(m.attributes['file']['path']))
            api.send(outports[2]['name'], m)


    api.send(outports[0]['name'], log_stream.getvalue())


inports = [{'name': 'input', 'type': 'message.file',"description":"Filename"}]
outports = [{'name': 'log', 'type': 'string',"description":"Logging data"},\
            {'name': 'numfiles', 'type': 'int64',"description":"Number of files"},\
            {'name': 'output', 'type': 'message.file',"description":"Filename"}]


#api.set_port_callback(inports[0]['name'], process)

def test_operator() :
    api.config.debug_mode = True
    api.config.start_date = '2020-02-01'
    api.config.end_date = '2020-03-31'

    process(api.Message(attributes={'file':{'path':'/folder/file-20200206'},'message.lastBatch':False},body = []))
    process(api.Message(attributes={'file':{'path':'/folder/file-20200207'},'message.lastBatch':False},body = []))
    process(api.Message(attributes={'file':{'path':'/folder/file-20200515'},'message.lastBatch':False},body = []))
    process(api.Message(attributes={'file': {'path': '/folder/file-20201215'}, 'message.lastBatch': True}, body=[]))

if __name__ == '__main__':
    test_operator()


    if True :
        subprocess.run(["rm", '-r',
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators' + api.config.version])
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name + '_' + api.config.version
        subprocess.run(["vctl", "solution", "bundle",
                        '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_' + api.config.version, \
                        "-t", solution_name])
        subprocess.run(["mv", solution_name + '.zip', '../../../solution/operators'])



import json
import re
import logging

tflogger = logging.getLogger('textfield_parser')

#### split text with quotes
# 1. find quote indices and separator indices
# 2. ignore all separator indices within quotes (only even number of quote indices less than sep index)
def split_text(text,sep) :
    sep_indices = [m.start() for m in re.finditer(sep, text)]
    quote_indices = [m.start() for m in re.finditer("'", text) if m.start() == 0 or text[m.start()-1] != '\\' ]

    list_seps = list()
    for s_index in sep_indices :
        if not len([ mi for mi in quote_indices if mi < s_index]) % 2 :
            list_seps.append(s_index)

    list_texts = list()
    last_index = 0
    for i,s_index in enumerate(list_seps) :
        if last_index == 0 :
            list_texts.append(text[:s_index].strip())
        else :
            list_texts.append(text[last_index+1:s_index].strip())
        last_index = s_index
    if last_index == 0 :
        list_texts.append(text[:].strip())
    else:
        list_texts.append(text[last_index + 1:].strip())

    return list_texts

#### READ Value
def read_value(text,test_number = True):
    if not text or text.upper() == 'NONE':
        return None
    val = text.strip().strip("'").strip('"')
    if test_number :
        val = number_test(val)
    tflogger.debug('Text -> Result: {} -> {} '.format(text,val))
    return  val

#### READ LIST
def read_list(text,value_list=None,sep = ',',modifier_list_not=None,test_number = True, ignore_quotes = False):

    if not text or text.upper() == 'NONE':
        return None

    if not modifier_list_not:
        modifier_list_not = ['!', '~', 'not', 'Not', 'NOT']

    #text = text.strip()

    # test for all
    if len(text) < 4 and ('all' in text or 'All' in text or 'ALL' in text) :
        return value_list

    negation = False
    # Test for Not
    if len(text) > 1 and text[0] in modifier_list_not :
        text = text[1:]
        negation = True
    elif len(text) > 3 and text[:3] in modifier_list_not :
        text = text[4:].strip()
        negation = True

    result_list = list()
    #elem_list = text.split(sep)
    elem_list = split_text(text, sep)
    for x in  elem_list:
        if ignore_quotes :
            elem = x.strip()
        else:
            elem = x.strip().strip("'\"")
        if test_number :
            elem = number_test(elem)
        result_list.append(elem)

    if negation :
        if not isinstance(value_list,list) :
            value_list = list(value_list)
        if not value_list  :
            raise ValueError("Negation needs a value list to exclude items")
        result_list = [x for x in value_list if x not in result_list]

    tflogger.debug('Text -> Result: {} -> {} '.format(text, result_list))
    return result_list

#### READ VALUE LIST
def read_dict_of_list(text,inner_sep = ',',outer_sep = ';',map_sep=':',test_number = True):
    if not text or text.upper() == 'NONE':
        return None

    vo_lists = [x.strip().strip("'\"") for x in text.strip().split(outer_sep)]

    value_list_dict = dict()
    for vo in vo_lists :
        key, vi_str = vo.split(map_sep)
        key = key.strip().strip("'\"")
        vi_list = vi_str.split(inner_sep)
        mapped_list = list()
        for vi in vi_list :
            elem = vi.strip().strip("'\"")
            if test_number :
                elem = number_test(elem)
            mapped_list.append(elem)
        value_list_dict[key] = mapped_list

    return value_list_dict


#### READ DICT

def read_dict(text, sep =',', map_sep=':', test_number = True, ignore_quotes = False):

    # replace \\' with \' and \\" with \"

    if not text or text.upper() == 'NONE':
        return None

    list_maps = split_text(text,sep)

    #list_maps = [x.strip() for x in text.split(sep)]
    map_dict = dict()
    for cm in list_maps :
        if ignore_quotes :
            key = cm.split(map_sep)[0].strip()
            value = cm.split(map_sep)[1].strip()
        else :
            mapping = split_text(cm,map_sep)
            key = mapping[0].strip().strip("'\"")
            value = mapping[1].strip().strip("'\"")
        if test_number :
            key = number_test(key)
            value = number_test(value)
        map_dict[key] = value
    return map_dict

#### READ DICT of DICT
# map_values : column1: {from_value: to_value}, column2: {from_value: to_value}
# 1. split outer_sep
# 2. split first occurrance of map_sep from rest - o_key
# 3. split rest by sep -> dict
def read_dict_of_dict(text, sep=',', map_sep=':', outer_sep = ';', test_number = True) :
    if not text or text.upper() == 'NONE':
        return None
    #if not outer_sep in text :  Produces Errors
    #    outer_sep = sep
    o_dict = dict()
    o_list = [x.strip() for x in text.split(outer_sep)]
    for ov in o_list:
        o_key, i_map = [x.strip() for x in ov.split(map_sep,1)]
        o_key = o_key.strip().strip("'\"")
        i_list = [x.strip() for x in i_map.split(sep)]
        i_dict = dict()
        for iv in i_list :
            i_key,value = [x.strip() for x in iv.split(map_sep)]
            i_key = i_key.strip().strip("(){}").strip().strip("'\"")
            value = value.strip().strip("(){}").strip().strip("'\"")
            if value.upper() == 'NAN' or value.upper() == 'NULL' or value.upper() == 'NONE':
                value = None
            if test_number :
                i_key = number_test(i_key)
                value = number_test(value)
            i_dict[i_key] = value
        o_dict[o_key] = i_dict
    return o_dict

#### READ JSON
def read_json(text) :
    if not text or text.upper() == 'NONE':
        return None
    j = json.loads(text)
    return j

#### READ Relation
def read_relations(text,sep=',',relation_map=None):
    if not text or text.upper() == 'NONE':
        return None
    if not relation_map :
        relation_map = {'!=': '!', '~':'!', '==': '=', '>=': '≥', '=>': '≥', '<=': '≤', '=<': '≤'}

    for key,value in relation_map.items() :
        text = text.replace(key,value)
    text_list = text.split(sep)

    print (text_list)

    relation_list = list()
    for selection in text_list:
        m = re.match(u'(.+)\s*([<>!≤≥=]+)\s*(.+)', selection)
        if m:
            left = m.group(1).strip().strip('"').strip("'")
            right = float(m.group(3).strip())
            relation = m.group(2).strip()
            relation_list.append([left,relation,right])
        else:
            raise ValueError('Could not parse relation statement: ' + selection)
    return relation_list

def number_test(str) :
    if str :
        if re.match('True',str) :
            return True
        elif re.match('False', str):
            return False
        elif re.match('None',str) :
            return None
        elif str.isdigit() :
            return int(str)
        else :
            try :
                f = float(str)
                return f
            except :
                return str
    else :
        return str

##############################
### MAIN
##############################
if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    #tflogger.setLevel(logging.DEBUG)
    tflogger.info('Main')
    ### list

    text_1 = "'1'"

    text = "'Hello', 'a list', with , 5, 5.6, separated by , me"
    not_text = "Not Mercedes, Renault, Citroen, Peugeaut, 'Rolls Royce'"
    list2 = ['Mercedes', 'Audi', 'VW', 'Skoda', 'Renault', 'Citroen', 'Peugeot', 'Rolls Royce']
    list_text3 = "'Mercedes', 'Audi, Volkswagen', 'VW, Volkswagen', 'Skoda, Volkswagen', 'Renault', 'Citroen, PSA', 'Peugeot', 'Rolls Royce'"
    print('Value: ' + str(read_value(text_1)))
    print('Value Number: ' + str(read_value(text_1,test_number=True)))
    print('Not: ' + str(read_list(not_text, list2)))
    print('None :' + str(read_list('')))
    print('None (literally):' + str(read_list('None')))
    print('List: ' + str(read_list(text)))

    print('ALL: ',str(read_list('ALL',list2)))

    vlist = read_list(list_text3)
    print('List with quotes and separators in quotes: \n{} -> \n{}'.format(list_text3,vlist))

    ### value lists
    value_str = "'Mercedes':expensive, German, respectable; Audi:'sportive, German, technology-advanced'; \
                    VW : 'people', 9,'solid', No1; Citroen:cool, Fantomas, CV2, elastic ; 'Rolls Rocye': royal,  \
                    British, 'very expensive', black, 8, True"

    print('Value lists: ' + str(read_dict_of_list(value_str)))

    ### map
    maplist = r"Mercedes:expensive, Audi:'\'sportive, \'expensive', VW : 'people, humble: except', Citroen:cool, 'Rolls Rocye': 'royal: pretended', 'Cars':6, 'eco':'False'"
    print('\nMap: \ninput: {}\nouput: {}\n'.format(maplist,read_dict(maplist)))

    maplist = "Mercedes:expensive, Audi:sportive, VW : people"
    print('\nMap: \ninput: {}\nouput: {}\n'.format(maplist,read_dict(maplist)))

    ### maps of maps
    dictdict = "'High Class':{'Mercedes':expensive, 'BWM':'sporty};'Sport Class':{ Porsche:'None',Ferrari:special}; \
                'Middle Class':{VW : 'people','Renault':'fashionable',Citroen:'classy, 'Peugeot':'modern'};\
                    'Luxury Class':{'Rolls Rocye': royal, Bentley:'rare'}; 'All:{4:8.9,6:90,7:7}"
    print('Maps of maps :' + str(read_dict_of_dict(dictdict)))


    ### json
    json_text = "{\"Luxury Class\": {\"Mercedes\":\"expensive\",\"Rolls Rocye\": \"royal\"}, \"High Middle Class\":{\"Audi\":\"sportiv\"}, \"Middle Class\": {}, \
                \"Middle Class\" : {\"Citroen\":\"cool\",\"VW\" : \"people\" }}"
    print('JSON: ' + str(read_json(json_text)))


    ### comparison
    comparison = ' anna > 1.70, norbert != 900, cindy <= 1.65'
    print('Comparison: ' + str(read_relations(comparison)))





import random
import sys
import time
from datetime import date
import json


MAX = 200

def main():

    for i in range(0,MAX) :
        dict_out= {"website":"test.com", "url":"http://xx.com", "date":str(date.today()), "title":"TITLE Something", "index":random.randint(0,MAX), "text":"There are a number of ways you can take to get the current date. We will use the date class of the datetime module to accomplish this task"}
        print(json.dumps(dict_out))
        if i%3 == 0 :
            sys.stderr.write("ERROR: {}: Mod 3".format(i)+"\n")
            time.sleep(0.1)
        time.sleep(0.1)


if __name__ == "__main__":
    main()
    # gs.gensolution(os.path.realpath(__file__), config, inports, outports)




import json
import os
import subprocess

import sdi_utils.gensolution as gs
import deprecated.set_logging as slog

try:
    api
except NameError:
    class api:
        class Message:
            def __init__(self,body = None,attributes = ""):
                self.body = body
                self.attributes = attributes
                
        def send(port,msg) :
            if port == outports[1]['name'] :
                print(msg.body)

        def set_config(config):
            api.config = config

        class config:
            ## Meta data
            config_params = dict()
            tags = {'sdi_utils':''}
            version = "0.0.1"
            operator_description = "dict to JSON-string"
            operator_name = 'json_dict'
            operator_description_long = "Converts dict to JSON-string"
            add_readme = dict()
            debug_mode = True
            config_params['debug_mode'] = {'title': 'Debug mode',
                                           'description': 'Sending debug level information to log port',
                                           'type': 'boolean'}


def process(msg) :
    att_dict = msg.attributes

    att_dict['operator'] = 'json_dict'
    logger, log_stream = slog.set_logging(att_dict['operator'], loglevel=api.config.debug_mode)
    logger.info("Process started. Logging level: {}".format(logger.level))

    logger.debug('Attributes: {}'.format(str(msg.attributes)))

    jstr = json.dumps(msg.body)

    api.send(outports[0]['name'], log_stream.getvalue())
    api.send(outports[1]['name'], api.Message(attributes=att_dict,body=jstr))


inports = [{'name': 'dict', 'type': 'message',"description":"Input dict"}]
outports = [{'name': 'log', 'type': 'string',"description":"Logging data"}, \
            {'name': 'json', 'type': 'message',"description":"Output JSON-string"}]


#api.set_port_callback(inports[0]['name'], process)

def test_operator() :
    config = api.config
    config.debug_mode = True
    api.set_config(config)

    data = [{'name': 'Anna', 'age': 25, 'born': '1995-03-14'}, {'name': 'Berta', 'age': 31, 'born': '1989-08-14'},
            {'name': 'Cecilia', 'age': 41, 'born': '1979-08-14'}]
    msg = api.Message(attributes={'format': 'json'}, body=data)
    process(msg)


if __name__ == '__main__':
    #test_operator()
    if True :
        gs.gensolution(os.path.realpath(__file__), api.config, inports, outports)
        solution_name = api.config.operator_name+'_'+api.config.version
        print('Solution name: {}'.format(solution_name))
        print('Current directory: {}'.format(os.getcwd()))
        subprocess.run(["vctl", "solution", "bundle", '/Users/d051079/OneDrive - SAP SE/GitHub/sdi_utils/solution/operators/sdi_utils_operators_0.0.1',\
                                  "-t", solution_name])
        subprocess.run(["mv", solution_name+'.zip', '../../../solution/operators'])





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sdi_utils",

    version="0.0.85",

    author="Thorsten Hapke",
    author_email="thorsten.hapke@sap.com",
    description="List of SAP DI helper functions like gensolution (package locally developed operators,textfield_parser, time_monitoring",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/thhapke/gensolution",
    keywords = ['SAP Data Intelligence','genjson','tprogress','textfield_parser'],
    packages=setuptools.find_packages(),
    install_requires=[
        'requests'
    ],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'gensolution = sdi_utils.gensolution:main'
        ]
    },
    classifiers=[
    	'Programming Language :: Python :: 3.5',
    	'Programming Language :: Python :: 3.6',
    	'Programming Language :: Python :: 3.7',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import time
import logging

# logs time progress of program
plogger = logging.getLogger('tprogress')
plogger.setLevel(logging.INFO)


class progress :

    def __init__(self,max_num=1,step_size=1) :
        self.inc = -1
        self.step_size = step_size
        self.max = max_num
        self.time_start = time.time()
        self.time_last = self.time_start
        self.time_end_projected = time.time()
        self.strtime_end_projected = time.strftime("%H:%M:%S", time.localtime(self.time_end_projected))
        plogger.debug("Start time: " + time.strftime('%H:%M:%S  %Y-%m-%d '))

    # called within a loop to state past time and projected end time
    def monitor(self):
        self.inc += 1
        if self.inc % self.step_size == 0:
            time_taken = time.time() - self.time_start
            if self.inc != 0 :
                self.time_end_projected = time_taken / (self.inc*self.step_size) * self.max + self.time_start
                self.strtime_end_projected = time.strftime("%H:%M:%S", time.localtime(self.time_end_projected))
            time_hour, time_min, time_sec = self.timestamp(time_taken)
            prog = int(round((self.inc*self.step_size*100)/self.max))
            time_str = "{0:3d}% : {1:3.0f} s ({2:3.0f}:{3:2.0f}:{4:3.1f}) End:{5}".format(prog,time_taken,time_hour, time_min, time_sec,self.strtime_end_projected)
            plogger.info("Monitor: " + time_str)

    # called to get time since last called
    def elapsed_time (self,stage='') :
        stage = stage.ljust(12)
        time_hour, time_min, time_sec = self.timestamp(time.time() - self.time_start)
        _, takenl_m, takenl_s = self.timestamp(time.time() - self.time_last,hour=False)
        tstr = "{:3.0f}m {:.3f}s ({:3.0f}:{:2.0f}:{:3.1f})".format(takenl_m, takenl_s, time_hour, time_min, time_sec)
        plogger.debug('Elapsed Time: ' + tstr + ' - ' + stage)
        self.time_last = time.time()
        return tstr

    def timestamp(self,time_s,hour=True):
        time_hour = 0
        if hour :
            time_hour, time_s = divmod(time_s, 3600)
        time_min, time_sec = divmod(time_s, 60)
        return time_hour, time_min, time_sec

    def get_start_time(self):
        return time.strftime("%H:%M:%S", time.localtime(self.time_start))